import numpy as np
import xarray as xr
from numpy.testing import assert_allclose

import xyzpy as xyz

from . import (
    foo2_array,
    foo2_array_array,
    foo2_zarray1_zarray2,
    foo3_float_bool,
    foo3_scalar,
)

# --------------------------------------------------------------------------- #
# CASE_RUNNER tests                                                           #
# --------------------------------------------------------------------------- #


class TestCaseRunner:
    def test_seq(self):
        cases = ((1, 10, 100), (2, 20, 200), (3, 30, 300))
        xs = xyz.case_runner(foo3_scalar, ("a", "b", "c"), cases, verbosity=0)
        assert xs == (111, 222, 333)

    def test_progbar(self):
        cases = ((1, 10, 100), (2, 20, 200), (3, 30, 300))
        xs = xyz.case_runner(foo3_scalar, ("a", "b", "c"), cases, verbosity=2)
        assert xs == (111, 222, 333)

    def test_constants(self):
        cases = ((1,), (2,), (3,))
        xs = xyz.case_runner(
            foo3_scalar, ("a", "b", "c"), cases, constants={"b": 10, "c": 100}
        )
        assert xs == (111, 112, 113)

    def test_parallel(self):
        cases = ((1, 10, 100), (2, 20, 200), (3, 30, 300))
        xs = xyz.case_runner(
            foo3_scalar, ("a", "b", "c"), cases, num_workers=1
        )
        assert xs == (111, 222, 333)

    def test_split(self):
        cases = ((1, 10, 100), (2, 20, 200), (3, 30, 300))
        a, b = xyz.case_runner(
            foo3_float_bool, ("a", "b", "c"), cases, split=True
        )
        assert a == (111, 222, 333)
        assert b == (False, True, False)

    def test_single_args(self):
        cases = (1, 2, 3)
        xs = xyz.case_runner(
            foo3_scalar, "a", cases, constants={"b": 10, "c": 100}
        )
        assert xs == (111, 112, 113)


class TestCaseRunnerToDS:
    def test_single(self):
        cases = [(1, 20, 300), (3, 20, 100)]
        ds = xyz.case_runner_to_ds(
            foo3_scalar, ["a", "b", "c"], cases=cases, var_names="sum"
        )
        assert_allclose(ds["a"].data, [1, 3])
        assert_allclose(ds["b"].data, [20])
        assert_allclose(ds["c"].data, [100, 300])
        assert ds["sum"].loc[{"a": 1, "b": 20, "c": 300}].data == 321
        assert ds["sum"].loc[{"a": 3, "b": 20, "c": 100}].data == 123
        assert ds["sum"].loc[{"a": 1, "b": 20, "c": 100}].isnull()
        assert ds["sum"].loc[{"a": 3, "b": 20, "c": 300}].isnull()

    def test_single_dict_cases(self):
        cases = [{"a": 1, "b": 20, "c": 300}, {"a": 3, "b": 20, "c": 100}]
        ds = xyz.case_runner_to_ds(
            foo3_scalar, None, cases=cases, var_names="sum"
        )
        assert_allclose(ds["a"].data, [1, 3])
        assert_allclose(ds["b"].data, [20])
        assert_allclose(ds["c"].data, [100, 300])
        assert ds["sum"].loc[{"a": 1, "b": 20, "c": 300}].data == 321
        assert ds["sum"].loc[{"a": 3, "b": 20, "c": 100}].data == 123
        assert ds["sum"].loc[{"a": 1, "b": 20, "c": 100}].isnull()
        assert ds["sum"].loc[{"a": 3, "b": 20, "c": 300}].isnull()

    def test_multires(self):
        cases = [(1, 20, 300), (3, 20, 100)]
        ds = xyz.case_runner_to_ds(
            foo3_float_bool,
            fn_args=["a", "b", "c"],
            cases=cases,
            var_names=["sum", "a_even"],
        )
        assert_allclose(ds["a"].data, [1, 3])
        assert_allclose(ds["b"].data, [20])
        assert_allclose(ds["c"].data, [100, 300])
        assert ds["sum"].loc[{"a": 1, "b": 20, "c": 300}].data == 321
        assert ds["sum"].loc[{"a": 3, "b": 20, "c": 100}].data == 123
        assert ds["sum"].loc[{"a": 1, "b": 20, "c": 100}].isnull()
        assert ds["sum"].loc[{"a": 3, "b": 20, "c": 300}].isnull()
        assert ds["a_even"].data.dtype == object
        assert bool(ds["a_even"].sel(a=1, b=20, c=300).data) is False
        assert bool(ds["a_even"].sel(a=3, b=20, c=100).data) is False
        assert ds["a_even"].loc[{"a": 1, "b": 20, "c": 100}].isnull()
        assert ds["a_even"].loc[{"a": 3, "b": 20, "c": 300}].isnull()

    def test_array_return(self):
        ds = xyz.case_runner_to_ds(
            fn=foo2_array,
            fn_args=["a", "b"],
            cases=[(2, 30), (4, 50)],
            var_names="x",
            var_dims=["time"],
            var_coords={"time": np.arange(10) / 10},
        )
        assert ds.x.data.dtype == float
        assert ds.x.sel(a=2, b=50, time=0.7).isnull()
        assert ds.x.sel(a=4, b=50, time=0.3).data == 54.3

    def test_multi_array_return(self):
        ds = xyz.case_runner_to_ds(
            fn=foo2_array_array,
            fn_args=["a", "b"],
            cases=[(2, 30), (4, 50)],
            var_names=["x", "y"],
            var_dims={("x", "y"): "time"},
            var_coords={"time": ["a", "b", "c", "d", "e"]},
        )
        assert ds["time"].data.dtype != object
        assert_allclose(ds["x"].sel(a=4, b=50).data, [50, 54, 58, 62, 66])
        assert_allclose(ds["y"].sel(a=4, b=50).data, [50, 46, 42, 38, 34])

    def test_align_and_fillna_int(self):
        ds1 = xyz.case_runner_to_ds(
            foo2_array_array,
            fn_args=["a", "b"],
            cases=[(1, 10), (2, 20)],
            var_names=["x", "y"],
            var_dims={("x", "y"): "time"},
            var_coords={"time": ["a", "b", "c", "d", "e"]},
        )
        ds2 = xyz.case_runner_to_ds(
            foo2_array_array,
            fn_args=["a", "b"],
            cases=[(2, 10), (1, 20)],
            var_names=["x", "y"],
            var_dims={("x", "y"): "time"},
            var_coords={"time": ["a", "b", "c", "d", "e"]},
        )
        assert not np.logical_not(ds1["x"].isnull()).all()
        assert not np.logical_not(ds1["y"].isnull()).all()
        assert not np.logical_not(ds2["x"].isnull()).all()
        assert not np.logical_not(ds2["y"].isnull()).all()
        ds1, ds2 = xr.align(ds1, ds2, join="outer")
        fds = ds1.fillna(ds2)
        assert np.logical_not(fds["x"].isnull()).all()
        assert np.logical_not(fds["y"].isnull()).all()

    def test_align_and_fillna_complex(self):
        ds1 = xyz.case_runner_to_ds(
            foo2_zarray1_zarray2,
            fn_args=["a", "b"],
            cases=[(1j, 10), (2j, 20)],
            var_names=["x", "y"],
            var_dims={("x", "y"): "time"},
            var_coords={"time": ["a", "b", "c", "d", "e"]},
        )
        ds2 = xyz.case_runner_to_ds(
            foo2_zarray1_zarray2,
            fn_args=["a", "b"],
            cases=[(2j, 10), (1j, 20)],
            var_names=["x", "y"],
            var_dims={("x", "y"): "time"},
            var_coords={"time": ["a", "b", "c", "d", "e"]},
        )
        assert not np.logical_not(np.isnan(ds1["x"].data)).all()
        assert not np.logical_not(np.isnan(ds1["y"].data)).all()
        assert not np.logical_not(np.isnan(ds2["x"].data)).all()
        assert not np.logical_not(np.isnan(ds2["y"].data)).all()
        assert all(
            t == "complex"
            for t in (ds1.x.dtype, ds2.x.dtype, ds1.y.dtype, ds2.y.dtype)
        )
        assert ds1.y.dtype == complex
        assert ds2.y.dtype == complex
        ds1, ds2 = xr.align(ds1, ds2, join="outer")
        fds = ds1.fillna(ds2)
        assert np.logical_not(np.isnan(fds["x"].data)).all()
        assert np.logical_not(np.isnan(fds["y"].data)).all()


class TestCaseRunnerToDF:
    def test_single_arg(self):

        df = xyz.case_runner_to_df(
            fn=foo3_scalar,
            fn_args=["a", "b", "c"],
            cases=[(1, 10, 100), (2, 20, 200)],
            var_names="x",
        )
        assert len(df) == 2
        assert df.columns.tolist() == ["a", "b", "c", "x"]
        assert "int" in df.x.dtype.name


# --------------------------------------------------------------------------- #
# Finding and filling missing data                                            #
# --------------------------------------------------------------------------- #


class TestFindMissingCases:
    def test_parse_into_cases(self):
        import random

        @xyz.label(None, harvester=True)
        def func(x, y, opt1, opt2, seed):
            return {
                "sum": x + y,
                "long": (("i",), np.linspace(x, y, 4)),
                "diff": int(x - y),
                "label": f"{opt1}-{opt2}-{seed}",
            }

        # generate initial dataset
        cases = [
            {"opt1": "hello", "opt2": "world"},
            {"opt1": "goodbye", "opt2": "everyone"},
        ]
        combos = dict(
            x=[1, 2, 3],
            y=np.linspace(10, 100, 20),
            seed=range(5),
        )

        cases = xyz.parse_into_cases(combos=combos, cases=cases)
        # randomly drop half
        random.shuffle(cases)
        cases = cases[: len(cases) // 2]
        func.harvest_combos(combos=None, cases=cases)

        # now test both filling in and adding new coordinates
        cases = [
            {"opt1": "hello", "opt2": "world"},
            {"opt1": "goodbye", "opt2": "world"},
        ]
        combos = dict(
            x=[2, 3, 4],
            y=np.linspace(10, 100, 20),
            seed=range(6),
        )

        cases_expected = xyz.parse_into_cases(
            combos=combos,
            cases=cases,
            ds=func.full_ds,
        )
        len(cases_expected)

        for case in cases_expected:
            assert xyz.is_case_missing(func.full_ds, case)

    def test_simple(self):
        ds = xr.Dataset(coords={"a": [1, 2, 3], "b": [40, 50]})
        ds["x"] = (
            ("a", "b"),
            np.array([[0.1, np.nan], [np.nan, 0.2], [np.nan, np.nan]]),
        )
        # Target cases and settings
        t_cases = ((1, 50), (2, 40), (3, 40), (3, 50))
        t_configs = tuple(dict(zip(["a", "b"], t_case)) for t_case in t_cases)
        # Missing cases and settings
        m_configs = xyz.find_missing_cases(ds)
        # Assert same set of coordinates
        assert all(t_config in m_configs for t_config in t_configs)
        assert all(m_config in t_configs for m_config in m_configs)

    def test_multires(self):
        ds = xr.Dataset(coords={"a": [1, 2, 3], "b": [40, 50]})
        ds["x"] = (
            ("a", "b"),
            np.array([[0.1, np.nan], [np.nan, 0.2], [np.nan, np.nan]]),
        )
        ds["y"] = (
            ("a", "b"),
            np.array([["a", None], [None, "b"], [None, None]]),
        )
        # Target cases and settings
        t_cases = ((1, 50), (2, 40), (3, 40), (3, 50))
        t_configs = tuple(dict(zip(["a", "b"], t_case)) for t_case in t_cases)
        # Missing cases and settings
        m_configs = xyz.find_missing_cases(ds)
        # Assert same set of coordinates
        m_args = {arg for m_config in m_configs for arg in m_config}
        assert set(m_args) == {"a", "b"}
        assert all(t_config in m_configs for t_config in t_configs)
        assert all(m_config in t_configs for m_config in m_configs)

    def test_var_dims_leave(self):
        ds = xr.Dataset(
            coords={"a": [1, 2, 3], "b": [40, 50], "t": [0.1, 0.2, 0.3]}
        )
        ds["x"] = (
            ("a", "b"),
            np.array([[0.1, np.nan], [np.nan, 0.2], [np.nan, np.nan]]),
        )
        ds["y"] = (
            ("a", "b", "t"),
            np.array(
                [
                    [[0.2] * 3, [np.nan] * 3],
                    [[np.nan] * 3, [0.4] * 3],
                    [[np.nan] * 3, [np.nan] * 3],
                ]
            ),
        )
        # Target cases and settings
        t_cases = ((1, 50), (2, 40), (3, 40), (3, 50))
        t_configs = tuple(dict(zip(["a", "b"], t_case)) for t_case in t_cases)
        # Missing cases and settings
        m_configs = xyz.find_missing_cases(ds, ignore_dims="t")
        m_args = {arg for m_config in m_configs for arg in m_config}
        # Assert same set of coordinates
        assert set(m_args) == {"a", "b"}
        assert all(t_config in m_configs for t_config in t_configs)
        assert all(m_config in t_configs for m_config in m_configs)



"""Manage datasets --- loading, saving, merging etc."""

# TODO: add singlet dimensions (for all or given vars) ---------------------- #

from glob import glob
from pathlib import Path

import joblib
import numpy as np
import xarray as xr

_DEFAULT_FN_CACHE_PATH = "__xyz_cache__"


def cache_to_disk(fn=None, *, cachedir=_DEFAULT_FN_CACHE_PATH, **kwargs):
    """Cache this function to disk, using joblib."""
    mem = joblib.Memory(cachedir, verbose=0, **kwargs)

    # bare decorator
    if fn:
        return mem.cache(fn)

    # decorator called with kwargs
    else:

        def cache_to_disk_decorator(fn):
            return mem.cache(fn)

        return cache_to_disk_decorator


_engine_extensions = {
    "h5netcdf": ".h5",
    "netcdf4": ".nc",
    "joblib": ".dmp",
    "zarr": ".zarr",
}


def auto_add_extension(file_name, engine):
    """Ensure ``file_name`` has an extension matching ``engine``.

    Parameters
    ----------
    file_name : str
        File name to normalize.
    engine : {'h5netcdf', 'netcdf4', 'joblib', 'zarr'}
        Engine determining the extension.

    Returns
    -------
    str
        File name with an appropriate extension appended.
    """
    if not any(ext in file_name for ext in _engine_extensions.values()):
        extension = _engine_extensions[engine]
        file_name += extension
    return file_name


def save_ds(ds, file_name, engine="h5netcdf", **kwargs):
    """Saves a xarray dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        The dataset to save.
    file_name : str
        Name of the file to save to.
    engine : {'h5netcdf', 'netcdf4', 'joblib', 'zarr'}, optional
        Engine used to save file with.

    Returns
    -------
        None
    """
    file_name = auto_add_extension(file_name, engine)

    # Parse out 'bad' netcdf types (only attributes -> values not so important)
    if engine not in {"joblib", "zarr"}:
        for attr, val in ds.attrs.items():
            if callable(val):
                ds.attrs[attr] = str(val)
            if val is None:
                ds.attrs[attr] = "None"
            if val is True:
                ds.attrs[attr] = "True"
            if val is False:
                ds.attrs[attr] = "False"

    if engine == "joblib":
        joblib.dump(ds, file_name, **kwargs)
    elif engine == "zarr":
        ds.to_zarr(file_name, **kwargs)
    else:
        # officially: complex data dtypes are not valid for netcdf
        if any(np.iscomplexobj(v.values) for v in ds.variables.values()):
            kwargs.setdefault("invalid_netcdf", True)
        ds.to_netcdf(file_name, engine=engine, **kwargs)


def load_ds(
    file_name,
    engine="h5netcdf",
    load_to_mem=None,
    create_new=False,
    chunks=None,
    **kwargs,
):
    """Loads a xarray dataset. Basically ``xarray.open_dataset`` with some
    different defaults and convenient behaviour.

    Parameters
    ----------
    file_name : str
        Name of file to open.
    engine : {'h5netcdf', 'netcdf4', 'joblib', 'zarr'}, optional
        Engine used to load file.
    load_to_mem : bool, optional
        Ince opened, load from disk into memory. Defaults to ``True`` if
        ``chunks=None``.
    create_new : bool, optional
        If no file exists make a blank one.
    chunks : int or dict
        Passed to ``xarray.open_dataset`` so that data is stored using
        ``dask.array``.


    Returns
    -------
    ds : xarray.Dataset
        Loaded Dataset.
    """
    file_name = auto_add_extension(file_name, engine)
    file_path = Path(file_name)

    if not file_path.exists() and create_new:
        return xr.Dataset()

    if engine == "joblib":
        return joblib.load(file_path, **kwargs)

    if (load_to_mem is None) and (chunks is None):
        load_to_mem = True
    else:
        if load_to_mem and (chunks is not None):
            raise ValueError("``chunks`` redundant if ``load_to_mem`` given.")
        else:
            load_to_mem = False

    if engine == "zarr":
        ds = xr.open_zarr(file_path, chunks=chunks, **kwargs)

    else:
        opts = dict(engine=engine, chunks=chunks, **kwargs)

        try:
            ds = xr.open_dataset(file_path, **opts)
        except AttributeError as e1:
            if "object has no attribute" in str(e1) and engine == "h5netcdf":
                opts["engine"] = "netcdf4"
                ds = xr.open_dataset(file_path, **opts)
            else:
                raise e1

    if load_to_mem:
        ds.load()
        ds.close()

    return ds


def save_merge_ds(ds, fname, overwrite=None, **kwargs):
    """Save dataset ``ds``, but check for an existing dataset with that name
    first, and if it exists, merge the two before saving.

    Parameters
    ----------
    ds : xarray.Dataset
        The dataset to save.
    fname : str
        The file name.
    overwrite : {None, False, True}, optional
        How to merge the dataset with the existing dataset.

            - None: the datasets will be merged in there are no conflicts
            - False: data will be taken from old dataset if conflicting
            - True: data will be taken from new dataset if conflicting
    """

    # check for the existing file
    if Path(fname).exists():
        old_ds = load_ds(fname)
    else:
        old_ds = xr.Dataset()

    # merge the new and old
    if overwrite is True:
        new_ds = ds.combine_first(old_ds)
    elif overwrite is False:
        new_ds = old_ds.combine_first(ds)
    else:
        new_ds = xr.merge([old_ds, ds], join="outer")

    # avoid string truncation
    new_ds = new_ds.drop_encoding()

    # write to disk
    save_ds(new_ds, fname, **kwargs)


def trimna(obj):
    """Drop values across dims where all values are NaN.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        Object to trim.

    Returns
    -------
    same type as obj
        Trimmed object.
    """
    trimmed_obj = obj.copy()
    for d in obj.dims:
        trimmed_obj = trimmed_obj.dropna(d, how="all")
    return trimmed_obj


def sort_dims(ds):
    """Reorder variable dimensions to match ``ds.dims``. This is an inplace
    operation.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to reorder in place.

    Returns
    -------
    None
    """
    dim_list = tuple(ds.dims)

    for var in ds.data_vars:
        var_dims = (d for d in dim_list if d in ds[var].dims)
        ds[var] = ds[var].transpose(*var_dims)


def post_fix(ds, postfix):
    """Append ``"_{postfix}"`` to each data variable name.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to rename.
    postfix : str
        Suffix to append.

    Returns
    -------
    xarray.Dataset
        Renamed dataset.
    """
    return ds.rename({old: old + "_" + postfix for old in ds.data_vars})


def check_runs(obj, dim="run", var=None, sel=()):
    """Print out information about the range and any missing values for an
    integer dimension.

    Parameters
    ----------
        obj : xarray object
            Data to check.
        dim : str (optional)
            Dimension to check, defaults to 'run'.
        var : str (optional)
            Subselect this data variable first.
        sel : mapping (optional)
            Subselect these other coordinates first.
    """
    try:
        if sel:
            obj = obj.loc[sel]
        if var:
            obj = obj[var]

        obj = trimna(obj)
        obj = obj.dropna(dim, how="all")
        obj = obj[dim].values
    except KeyError:
        print("* NO DATA *")
        return

    if obj.size == 0:
        print("* NO DATA *")
        return

    if "int" not in str(obj.dtype):
        raise TypeError("check_runs can only check integer dimesions.")

    xmin, xmax = obj.min(), obj.max() + 1
    xmiss_start = obj[:-1][obj[1:] != obj[:-1] + 1] + 1
    xmiss_end = obj[1:][obj[1:] != obj[:-1] + 1]
    xmissing = list(zip(xmiss_start, xmiss_end))
    msg = "RANGE:  {:>7} -> {:<7} | TOTAL:  {:^7}".format(xmin, xmax, len(obj))
    if xmissing:
        msg += " | MISSING: {}".format(xmissing)
    print(msg)


def auto_xyz_ds(x, y_z=None):
    """Automatically turn an array into a `xarray` dataset. Transpose ``y_z``
    if necessary to automatically match dimension sizes.

    Parameters
    ----------
    x : array_like
        The x-coordinates.
    y_z : array_like, optional
        The y-data, possibly varying with coordinate z.
    """
    # assume non-coordinated - e.g. for auto_histogram
    if y_z is None:
        dims = ["yzwvu"[i] for i in range(x.ndim)]
        return xr.Dataset(data_vars={"x": (dims, x)})

    # Infer dimensions to coords mapping
    y_z = np.array(np.squeeze(y_z), ndmin=2)
    x = np.asarray(x)
    if np.size(x) == y_z.shape[0]:
        y_z = np.transpose(y_z)
    n_y = y_z.shape[0]

    # Turn into dataset
    if x.ndim == 2:
        ds = xr.Dataset(
            data_vars={"y": (["z", "_x"], y_z), "x": (["z", "_x"], x)}
        )
    else:
        ds = xr.Dataset(
            coords={"x": x, "z": np.arange(n_y)},
            data_vars={"y": (["z", "x"], y_z)},
        )
    return ds


def merge_sync_conflict_datasets(
    base_name, engine="h5netcdf", combine_first=False
):
    """Glob files based on `base_name`, merge them, save this new dataset if
    it contains new info, then clean up the conflicts.

    Parameters
    ----------
        base_name : str
            Base file name to glob on - should include '*'.
        engine : str , optional
            Load and save engine used by xarray.
        combine_first : bool, optional
            If True, combine datasets sequentially using ``combine_first``,
            preferring the first dataset in the list, which is assumed to be
            the original. If False, merge all datasets together using
            ``xr.merge``, which will raise an error if there are any conflicts.
    """
    fnames = glob(base_name)
    if len(fnames) < 2:
        print("Nothing to do - need multiple files to merge.")
        return

    # make sure first filename is the shortest -> assumed original
    fnames.sort(key=len)

    print("Merging:\n{}\ninto ->\n{}\n".format(fnames, fnames[0]))

    def load_dataset(fname):
        return load_ds(fname, engine=engine)

    datasets = list(map(load_dataset, fnames))

    # combine all the conflicts
    if combine_first:
        full_dataset = datasets[0]
        for ds in datasets[1:]:
            full_dataset = full_dataset.combine_first(ds)
    else:
        full_dataset = xr.merge(datasets, join="outer")

    # avoid string truncation
    full_dataset = full_dataset.drop_encoding()

    # save new dataset?
    if full_dataset.identical(datasets[0]):
        # nothing to do
        pass
    else:
        save_ds(full_dataset, fnames[0], engine=engine)

    # clean up conflicts
    for fname in fnames[1:]:
        Path(fname).unlink()


def save_df(df, name, engine="pickle", key="df", **kwargs):
    """Save a dataframe to disk.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame to save.
    name : str
        File name to save to.
    engine : {'pickle', 'csv', 'hdf'}, optional
        Storage backend.
    key : str, optional
        HDF key when ``engine='hdf'``.
    **kwargs
        Passed through to the pandas writer.
    """
    meth = "to_{}".format(engine)
    if engine == "hdf":
        kwargs["key"] = key
    if engine == "csv":
        kwargs.setdefault("index", False)
    getattr(df, meth)(name, **kwargs)


def load_df(name, engine="pickle", key="df", **kwargs):
    """Load a dataframe from disk.

    Parameters
    ----------
    name : str
        File name to read from.
    engine : {'pickle', 'csv', 'hdf'}, optional
        Storage backend.
    key : str, optional
        HDF key when ``engine='hdf'``.
    **kwargs
        Passed through to the pandas reader.

    Returns
    -------
    pandas.DataFrame
        Loaded dataframe.
    """
    import pandas as pd

    func = "read_{}".format(engine)
    if engine == "hdf":
        kwargs["key"] = key
    return getattr(pd, func)(name)



import os
import sys
from tempfile import TemporaryDirectory

import numpy as np
import pytest
import xarray as xr
from numpy.testing import assert_allclose

from xyzpy import Harvester, Runner, combo_runner, combo_runner_to_ds, label
from xyzpy.gen.cropping import (
    Crop,
    XYZError,
    _acquire_affinity,
    _acquire_gpu,
    _parse_resource_ids,
    _ResourcePool,
    grow,
    load_crops,
    parse_crop_details,
)

from . import (
    foo2_array,
    foo2_array_bool,
    foo2_dataset,
    foo2_scalar,
    foo3_scalar,
)


def foo_add(a, b, c):
    return a + b


class TestSowerReaper:
    @pytest.mark.parametrize(
        "fn, crop_name, crop_loc, expected",
        [
            (foo_add, None, None, ".xyz-foo_add"),
            (None, "custom", None, ".xyz-custom"),
            (foo_add, "custom", None, ".xyz-custom"),
            (
                foo_add,
                None,
                "custom_dir",
                os.path.join("custom_dir", ".xyz-foo_add"),
            ),
            (
                None,
                "custom",
                "custom_dir",
                os.path.join("custom_dir", ".xyz-custom"),
            ),
            (
                foo_add,
                "custom",
                "custom_dir",
                os.path.join("custom_dir", ".xyz-custom"),
            ),
            (None, None, None, "raises"),
            (None, None, "custom_dir", "raises"),
        ],
    )
    def test_parse_crop_details(self, fn, crop_name, crop_loc, expected):

        if expected == "raises":
            with pytest.raises(ValueError):
                parse_crop_details(fn, crop_name, crop_loc)
        else:
            crop_loc = parse_crop_details(fn, crop_name, crop_loc)[0]
            assert crop_loc[-len(expected) :] == expected

    def test_checks(self):
        with pytest.raises(ValueError):
            Crop(name="custom", save_fn=True)

        with pytest.raises(TypeError):
            c = Crop(fn=foo_add, save_fn=False, batchsize=0.5)
            c.choose_batch_settings(combos=[("a", [1, 2])])

        with pytest.raises(ValueError):
            c = Crop(fn=foo_add, save_fn=False, batchsize=-1)
            c.choose_batch_settings(combos=[("a", [1, 2])])

        with pytest.raises(ValueError):
            c = Crop(fn=foo_add, save_fn=False, batchsize=1, num_batches=2)
            c.choose_batch_settings(combos=[("a", [1, 2, 3])])

        with pytest.raises(ValueError):
            c = Crop(fn=foo_add, save_fn=False, batchsize=2, num_batches=3)
            c.choose_batch_settings(combos=[("a", [1, 2, 3])])

        with pytest.raises(TypeError):
            c = Crop(fn=foo_add, save_fn=False, batchsize=1, num_batches="3")
            c.choose_batch_settings(combos=[("a", [1, 2, 3])])

        c = Crop(fn=foo_add, save_fn=False, batchsize=1, num_batches=3)
        c.choose_batch_settings(combos=[("a", [1, 2, 3])])

        c = Crop(fn=foo_add, save_fn=False, batchsize=2, num_batches=2)
        c.choose_batch_settings(combos=[("a", [1, 2, 3])])

        c = Crop(fn=foo_add, save_fn=False, batchsize=3, num_batches=1)
        c.choose_batch_settings(combos=[("a", [1, 2, 3])])

        with pytest.raises(XYZError):
            grow(1)

        print(c)
        repr(c)

    def test_choose_batch_settings_ignores_stale_batch_remainder(self):
        c = Crop(fn=foo_add, save_fn=False, batchsize=1, num_batches=3)
        c._batch_remainder = 1

        c.choose_batch_settings(combos=[("a", [1, 2, 3])])

        assert c._batch_remainder == 0

    @pytest.mark.parametrize("shuffle", [False, True, 2])
    def test_batch(self, shuffle):

        combos = [
            ("a", [10, 20, 30]),
            ("b", [4, 5, 6, 7]),
        ]
        expected = combo_runner(foo_add, combos, constants={"c": True})

        with TemporaryDirectory() as tdir:
            # sow seeds
            crop = Crop(fn=foo_add, parent_dir=tdir, batchsize=5)

            assert not crop.is_prepared()
            assert crop.num_sown_batches == crop.num_results == -1

            crop.sow_combos(combos, constants={"c": True}, shuffle=shuffle)

            assert crop.is_prepared()
            assert crop.num_sown_batches == 3
            assert crop.num_results == 0

            # grow seeds
            for i in range(1, 4):
                grow(i, Crop(parent_dir=tdir, name="foo_add"))

                if i == 1:
                    assert crop.missing_results() == (
                        2,
                        3,
                    )

                    with pytest.raises(XYZError):
                        crop.reap()

            assert crop.is_ready_to_reap()
            assert not crop.check_bad()
            # reap results
            results = crop.reap()

        assert results == expected

    def test_field_name_and_overlapping(self):
        combos1 = [("a", [10, 20, 30]), ("b", [4, 5, 6, 7])]
        expected1 = combo_runner(foo_add, combos1, constants={"c": True})

        combos2 = [("a", [40, 50, 60]), ("b", [4, 5, 6, 7])]
        expected2 = combo_runner(foo_add, combos2, constants={"c": True})

        with TemporaryDirectory() as tdir:
            # sow seeds
            c1 = Crop(name="run1", fn=foo_add, parent_dir=tdir, batchsize=5)
            c1.sow_combos(combos1, constants={"c": True})
            c2 = Crop(name="run2", fn=foo_add, parent_dir=tdir, batchsize=5)
            c2.sow_combos(combos2, constants={"c": True})

            # grow seeds
            for i in range(1, 4):
                grow(i, Crop(parent_dir=tdir, name="run1"))
                grow(i, Crop(parent_dir=tdir, name="run2"))

            # reap results
            assert not c1.check_bad()
            assert not c2.check_bad()
            results1 = c1.reap()
            results2 = c2.reap()

        assert results1 == expected1
        assert results2 == expected2

    @pytest.mark.parametrize("num_workers", [None, 2])
    @pytest.mark.parametrize("shuffle", [False, True, 2])
    def test_crop_grow_missing(self, num_workers, shuffle):
        combos1 = [("a", [10, 20, 30]), ("b", [4, 5, 6, 7])]
        expected1 = combo_runner(foo_add, combos1, constants={"c": True})
        with TemporaryDirectory() as tdir:
            c1 = Crop(name="run1", fn=foo_add, parent_dir=tdir, batchsize=5)
            c1.sow_combos(combos1, constants={"c": True}, shuffle=shuffle)
            c1.grow_missing(num_workers=num_workers)
            results1 = c1.reap()
        assert results1 == expected1

    def test_combo_reaper_to_ds(self):
        combos = (
            ("a", [1, 2]),
            ("b", [10, 20, 30]),
            ("c", [100, 200, 300, 400]),
        )

        with TemporaryDirectory() as tdir:
            # sow seeds
            crop = Crop(fn=foo3_scalar, parent_dir=tdir, batchsize=5)
            crop.sow_combos(combos)

            # check on disk repr works
            print(crop)
            repr(crop)

            # grow seeds
            for i in range(1, 6):
                crop.grow(i)

                if i == 3:
                    with pytest.raises(XYZError):
                        crop.reap_combos_to_ds(var_names=["bananas"])

            ds = crop.reap_combos_to_ds(var_names=["bananas"])

        assert ds.sel(a=2, b=30, c=400)["bananas"].data == 432

    @pytest.mark.parametrize("num_batches", [67, 98])
    def test_num_batches_doesnt_divide(self, num_batches):
        combos = (
            ("a", [1, 2, 3]),
            ("b", [10, 20, 30]),
            ("c", range(100, 1101, 100)),
        )

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir, num_batches=num_batches)
            crop.sow_combos(combos)
            assert crop.num_batches == num_batches
            crop.grow_missing()
            ds = crop.reap_combos_to_ds(var_names=["sum"])

        assert ds["sum"].sel(a=3, b=30, c=1100).data == 33

    @pytest.mark.parametrize(
        "fn",
        [
            foo2_scalar,
            foo2_array,
            foo2_array_bool,
            foo2_dataset,
        ],
    )
    def test_all_nan_result(self, fn):

        combos = (("a", [1, 2, 3]), ("b", [10, 20, 30]))

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=fn, parent_dir=tdir)
            crop.sow_combos(combos)

            with pytest.raises(XYZError):
                crop.all_nan_result

            crop.grow(1)
            nres = crop.all_nan_result

            if fn is foo2_array_bool:
                assert len(nres) == 2
                assert_allclose(nres[0], np.broadcast_to(np.nan, [10]))
                assert_allclose(nres[1], np.broadcast_to(np.nan, []))

            if fn is foo2_dataset:
                ds_exp = xr.Dataset(
                    {"x": (["t1", "t2"], np.tile(np.nan, (2, 3)))}
                )
                assert nres.identical(ds_exp)

    @pytest.mark.parametrize("shuffle", [False, True, 2])
    def test_reap_allow_incomplete(self, shuffle):
        combos = (
            ("a", [1, 2, 3]),
            ("b", [10, 20, 30]),
            ("c", range(100, 1101, 100)),
        )

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir)
            crop.sow_combos(combos, shuffle=shuffle)
            with pytest.raises(XYZError):
                crop.reap(allow_incomplete=True)
            for i in range(1, 40):
                crop.grow(i)
            res = np.array(crop.reap(allow_incomplete=True))
            assert np.isnan(res).sum() == 60

    def test_reap_allow_incomplete_with_short_final_batch(self):
        combos = (
            ("a", [1, 2]),
            ("b", [10, 20]),
            ("c", [100, 200]),
        )

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir, batchsize=3)
            crop.sow_combos(combos)

            assert crop.batchsize == 3
            assert crop.num_batches == 3
            assert crop._batch_remainder == 0

            crop.grow(1)
            crop.grow(2)

            res = np.array(crop.reap(allow_incomplete=True))

        assert res.shape == (2, 2, 2)
        assert np.isnan(res).sum() == 2

    @pytest.mark.parametrize(
        "fn,var_names,var_dims",
        [
            (foo2_scalar, ["x"], None),
            (foo2_array, ["x"], {"x": "t"}),
            (foo2_array_bool, ["x", "y"], {"x": "t"}),
            (foo2_dataset, None, None),
        ],
    )
    def test_reap_to_ds_allow_incomplete(self, fn, var_names, var_dims):
        combos = (("a", [1, 2, 3]), ("b", [10, 20, 30]))

        ds_exp = combo_runner_to_ds(fn, combos, var_names, var_dims=var_dims)

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=fn, parent_dir=tdir)
            crop.sow_combos(combos)
            for i in range(1, 10, 2):
                crop.grow(i)

            ds = crop.reap_combos_to_ds(
                var_names=var_names, var_dims=var_dims, allow_incomplete=True
            )

            num_finished = int(ds["x"].size * (5 / 9))
            assert (ds["x"] == ds_exp["x"]).sum() == num_finished

            crop.grow_missing()
            ds = crop.reap_combos_to_ds(
                var_names=var_names, var_dims=var_dims, allow_incomplete=True
            )
            assert ds.identical(ds_exp)

    def test_new_ds_crop_loads_info_incomplete(self):
        def fn(a, b):
            return xr.Dataset({"sum": a + b, "diff": a - b})

        with TemporaryDirectory() as tdir:
            disk_ds = os.path.join(tdir, "test.h5")

            combos = dict(a=[1], b=[1, 2, 3])
            runner = Runner(fn, var_names=None)
            harvester = Harvester(runner, disk_ds)
            crop = harvester.Crop(name="fn", batchsize=1, parent_dir=tdir)
            crop.sow_combos(combos)
            for i in range(1, 3):
                crop.grow(i)

            # try creating crop from fresh
            c = Crop(name="fn", parent_dir=tdir)
            # crop's harvester should be loaded from disk
            assert c.farmer is not None
            assert c.farmer is not harvester
            ds = c.reap(allow_incomplete=True)
            assert isinstance(ds, xr.Dataset)
            assert ds["diff"].isnull().sum() == 1
            assert harvester.full_ds["diff"].isnull().sum() == 1

            # try creating crop from harvester
            c = harvester.Crop("fn", parent_dir=tdir)
            # crop's harvester should still be harvester
            assert c.farmer is not None
            assert c.farmer is harvester
            ds = c.reap(allow_incomplete=True)
            assert isinstance(ds, xr.Dataset)
            assert ds["diff"].isnull().sum() == 1

    def test_load_crops(self):

        combos = (
            ("a", [1, 2, 3]),
            ("b", [10, 20, 30]),
            ("c", range(100, 1101, 100)),
        )

        with TemporaryDirectory() as tdir:
            c1 = Crop(name="Alice", fn=foo_add, parent_dir=tdir)
            c2 = Crop(name="Bob", fn=foo_add, parent_dir=tdir)

            c1.sow_combos(combos)
            c2.sow_combos(combos)

            crops = load_crops(tdir)
            assert "Alice" in crops
            assert "Bob" in crops
            assert len(crops) == 2

            c1.grow_missing()
            c2.grow_missing()

            assert (
                c1.reap_combos()
                == c2.reap_combos()
                == combo_runner(foo_add, combos)
            )

    @pytest.mark.parametrize("scheduler", ["sge", "pbs", "slurm"])
    def test_gen_cluster_script(self, scheduler):
        combos = [
            ("a", [10, 20, 30]),
            ("b", [4, 5, 6, 7]),
        ]

        with TemporaryDirectory() as tdir:
            # sow seeds
            crop = Crop(fn=foo_add, parent_dir=tdir)
            crop.sow_combos(combos, constants={"c": True})

            # test script to grow all
            s1 = crop.gen_cluster_script(scheduler=scheduler, minutes=20)
            print(s1)

            # test script to grow specified
            s2 = crop.gen_cluster_script(batch_ids=[0, 1], scheduler=scheduler)
            print(s2)

            # test script to grow missing
            crop.grow((1, 3, 5))
            s3 = crop.gen_cluster_script(scheduler=scheduler)
            print(s3)

            assert s1 != s2
            assert s1 != s3
            assert s2 != s3

    def test_sow_reap_cases(self):

        def dummy_function(a, b):
            return a + b

        cases1 = [(1, 2), (3, 4)]
        cases2 = [(5, 6), (7, 8)]
        fn_args = ("a", "b")

        with TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, "test.dmp")
            runner = Runner(dummy_function, var_names=["Dummy_Value"])
            harvester = Harvester(runner, engine="joblib", data_name=fl_pth)

            crop1 = harvester.Crop(parent_dir=tmpdir)
            crop1.sow_cases(fn_args=fn_args, cases=cases1)
            crop1.grow_missing()
            crop1.reap()

            crop2 = harvester.Crop(parent_dir=tmpdir)
            crop2.sow_cases(fn_args=fn_args, cases=cases2)
            crop2.grow_missing()
            crop2.reap()

            ds = harvester.full_ds
            assert isinstance(ds, xr.Dataset)
            assert ds["Dummy_Value"].size == 16
            assert ds["Dummy_Value"].notnull().sum() == 4

    def test_sow_reap_mixed_combos_cases(self):

        @label("out")
        def fn(a, b, c, d, e):
            return f"{a}{b}{c}{d}{e}"

        with TemporaryDirectory() as tmpdir:
            crop = fn.Crop("test", parent_dir=tmpdir)
            cases = [
                {"a": 1, "c": 3},
                {"a": 2, "c": 4},
            ]
            combos = {
                "b": [5, 6],
                "d": [7, 8],
            }
            constants = {"e": 9}
            crop.sow_combos(combos=combos, cases=cases, constants=constants)
            crop.grow_missing()
            ds = crop.reap()

        assert ds["out"].ndim == 4
        assert ds["out"].notnull().sum().item() == 8

    def test_delete_all_resets_state(self):
        combos = [
            ("a", [10, 20, 30]),
            ("b", [4, 5, 6, 7]),
        ]

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir, batchsize=5)
            crop.sow_combos(combos, constants={"c": True})

            # confirm crop is prepared and has loaded state
            assert crop.is_prepared()
            assert crop.batchsize == 5
            assert crop.num_batches == 3
            assert crop._batch_remainder is not None or True
            crop.calc_progress()
            assert crop._num_sown_batches == 3
            assert crop._num_results == 0

            # grow one batch to populate cached result state
            grow(1, Crop(parent_dir=tdir, name="foo_add"))
            crop.calc_progress()
            assert crop._num_results == 1

            crop.delete_all()

            # directory should be gone
            assert not os.path.exists(crop.location)

            # all loaded information should be reset
            assert crop.batchsize is None
            assert crop.num_batches is None
            assert crop._batch_remainder is None
            assert crop._all_nan_result is None
            assert crop._num_sown_batches == -1
            assert crop._num_results == -1

            # identity attributes should be preserved
            assert crop.name == "foo_add"
            assert crop.parent_dir is not None
            assert crop.location is not None
            assert crop.fn is foo_add


class TestParseResourceIds:
    def test_from_int(self):
        assert _parse_resource_ids(3) == [3]

    def test_from_list(self):
        assert _parse_resource_ids([0, 1, 2]) == [0, 1, 2]

    def test_from_tuple(self):
        assert _parse_resource_ids((4, 5)) == [4, 5]

    def test_from_range(self):
        assert _parse_resource_ids(range(3)) == [0, 1, 2]

    def test_from_str(self):
        assert _parse_resource_ids("0,1,2") == [0, 1, 2]

    def test_from_str_single(self):
        assert _parse_resource_ids("7") == [7]


class TestAcquireFunctions:
    def test_acquire_affinity(self):
        pargs = []
        env = {}
        _acquire_affinity(5, pargs, env)
        assert pargs == ["taskset", "-c", "5"]
        assert env == {}

    def test_acquire_gpu(self):
        pargs = []
        env = {}
        _acquire_gpu(2, pargs, env)
        assert pargs == []
        assert env == {"CUDA_VISIBLE_DEVICES": "2"}


class TestResourcePool:
    def test_from_raw_none(self):
        assert _ResourcePool.from_raw(None, _acquire_gpu) is None

    def test_from_raw_int(self):
        pool = _ResourcePool.from_raw(3, _acquire_gpu)
        assert pool.free == [3]

    def test_from_raw_str(self):
        pool = _ResourcePool.from_raw("0,1,2", _acquire_gpu)
        assert pool.free == [0, 1, 2]

    def test_from_raw_list(self):
        pool = _ResourcePool.from_raw([4, 5], _acquire_affinity)
        assert pool.free == [4, 5]

    def test_available_when_free(self):
        pool = _ResourcePool.from_raw([0, 1], _acquire_gpu)
        assert pool.available()

    def test_lifecycle(self):
        pool = _ResourcePool.from_raw("0,1,2", _acquire_gpu)
        assert pool.available()

        # acquire all three
        envs = {}
        for batch_id in range(3):
            pargs, env = [], {}
            pool.acquire(batch_id, pargs, env)
            envs[batch_id] = env

        assert not pool.available()
        assert len(pool.used) == 3
        # each got a different GPU (popped from end)
        assert envs[0] == {"CUDA_VISIBLE_DEVICES": "2"}
        assert envs[1] == {"CUDA_VISIBLE_DEVICES": "1"}
        assert envs[2] == {"CUDA_VISIBLE_DEVICES": "0"}

        # release middle one
        pool.release(1)
        assert pool.available()
        assert pool.free == [1]
        assert 1 not in pool.used

        # re-acquire
        pargs, env = [], {}
        pool.acquire(99, pargs, env)
        assert env == {"CUDA_VISIBLE_DEVICES": "1"}
        assert not pool.available()

    def test_affinity_acquire_prepends(self):
        pool = _ResourcePool.from_raw([7], _acquire_affinity)
        pargs = ["python", "-m", "my_module"]
        env = {}
        pool.acquire(0, pargs, env)
        assert pargs == ["taskset", "-c", "7", "python", "-m", "my_module"]
        assert env == {}


def foo_verbose(a, b, c):
    """Function that prints a lot of output to stdout."""
    for _ in range(1000):
        print("-" * 100)
    return a + b


class TestGrowSubprocessOutput:
    def test_grow_subprocess_heavy_stdout_no_hang(self):
        """Subprocess that prints >64KB to stdout should not deadlock."""
        combos = [("a", [1, 2]), ("b", [10, 20])]
        expected = combo_runner(foo_verbose, combos, constants={"c": True})

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_verbose, parent_dir=tdir, batchsize=2)
            crop.sow_combos(combos, constants={"c": True})
            crop.grow_subprocess(num_workers=1)
            assert crop.is_ready_to_reap()
            results = crop.reap()

        assert results == expected

    def test_grow_subprocess_save_logs(self):
        """log=True creates log files in the crop directory."""
        combos = [("a", [1, 2]), ("b", [10, 20])]
        expected = combo_runner(foo_verbose, combos, constants={"c": True})

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_verbose, parent_dir=tdir, batchsize=2)
            crop.sow_combos(combos, constants={"c": True})
            num_batches = crop.num_batches
            crop.grow_subprocess(num_workers=1, log=True)
            assert crop.is_ready_to_reap()

            # check log files before reap (reap cleans up the crop dir)
            log_dir = os.path.join(crop.location, "logs")
            assert os.path.isdir(log_dir)
            log_fnames = sorted(os.listdir(log_dir))
            assert len(log_fnames) == num_batches
            for fname in log_fnames:
                fpath = os.path.join(log_dir, fname)
                assert os.path.getsize(fpath) > 0

            results = crop.reap()

        assert results == expected

    def test_grow_subprocess_no_save_logs_no_log_dir(self):
        """log=False (default) should not create a logs directory."""
        combos = [("a", [1, 2]), ("b", [10])]

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir, batchsize=2)
            crop.sow_combos(combos, constants={"c": True})
            crop.grow_subprocess(num_workers=1)

            log_dir = os.path.join(crop.location, "logs")
            assert not os.path.exists(log_dir)


class TestGrowSubprocessResources:
    def test_grow_subprocess_with_gpus(self):
        """grow_subprocess with gpus= completes and produces results,
        even without real GPU hardware. CUDA_VISIBLE_DEVICES set on a
        pure-Python subprocess is harmless.
        """
        combos = [("a", [10, 20, 30]), ("b", [4, 5, 6, 7])]
        expected = combo_runner(foo_add, combos, constants={"c": True})

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir, batchsize=5)
            crop.sow_combos(combos, constants={"c": True})
            crop.grow_subprocess(
                num_workers=2,
                gpus="0,1",
            )
            assert crop.is_ready_to_reap()
            results = crop.reap()

        assert results == expected

    def test_grow_subprocess_with_gpus_single(self):
        """gpus= as a single int limits concurrency to 1."""
        combos = [("a", [1, 2]), ("b", [10, 20])]
        expected = combo_runner(foo_add, combos, constants={"c": True})

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir, batchsize=2)
            crop.sow_combos(combos, constants={"c": True})
            crop.grow_subprocess(
                num_workers=4,
                gpus=0,
            )
            assert crop.is_ready_to_reap()
            results = crop.reap()

        assert results == expected

    @pytest.mark.skipif(
        sys.platform != "linux",
        reason="taskset is only available on Linux",
    )
    def test_grow_subprocess_with_affinities_and_gpus(self):
        """Both affinities= and gpus= can be used together."""
        combos = [("a", [1, 2]), ("b", [10, 20])]
        expected = combo_runner(foo_add, combos, constants={"c": True})

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir, batchsize=2)
            crop.sow_combos(combos, constants={"c": True})
            crop.grow_subprocess(
                num_workers=4,
                affinities="0,1",
                gpus="0,1",
            )
            assert crop.is_ready_to_reap()
            results = crop.reap()

        assert results == expected



import argparse
import os
import sys
from pathlib import Path
from typing import Optional


def parse_num_workers(value: Optional[str]) -> Optional[int]:
    if value is None:
        return None
    try:
        return int(value)
    except ValueError:
        raise argparse.ArgumentTypeError("Invalid value for num_workers")


def parse_int_or_float(string):
    if string is None:
        return None
    try:
        return int(string)
    except ValueError:
        return float(string)


def parse_bool_flag(value: Optional[str] = None) -> bool:
    if value is None:
        return False
    if value.lower() in ("false", "0", "no"):
        return False
    elif value.lower() in ("true", "1", "yes"):
        return True
    else:
        raise argparse.ArgumentTypeError(f"Invalid boolean value {value}.")


def main():
    parser = argparse.ArgumentParser(
        description="Grow crops using xyzpy-gen-cropping."
    )
    parser.add_argument(
        "crop_name", type=str, help="The name of the crop to grow."
    )
    parser.add_argument(
        "--parent-dir",
        type=str,
        default=".",
        help="The parent directory of the crop.",
    )
    parser.add_argument(
        "--batch-ids",
        type=str,
        default="missing",
        help=(
            "Comma separated list of which batches to grow, "
            "by default all missing results."
        ),
    )
    parser.add_argument(
        "--raise-errors",
        nargs="?",
        const=True,
        default=False,
        type=parse_bool_flag,
        help="Raise batch errors.",
    )
    parser.add_argument(
        "--num-threads",
        type=int,
        default=1,
        help="The number of threads per worker (OMP_NUM_THREADS etc.)",
    )
    parser.add_argument(
        "--num-workers",
        type=parse_num_workers,
        default=None,
        help="The number of worker processes to use.",
    )
    parser.add_argument(
        "--subprocess",
        nargs="?",
        const=True,
        default=False,
        type=parse_bool_flag,
        help=(
            "Run each batch in its own fresh subprocess. This is most robust "
            "in terms of memory, at the cost of the process startup overhead. "
            "Optional value: true/false."
        ),
    )
    parser.add_argument(
        "--log",
        nargs="?",
        const=True,
        default=False,
        type=parse_bool_flag,
        help=(
            "Save subprocess stdout and stderr to log files in the crop "
            "directory under logs/batch-{batch_id}.log. Only used when "
            "--subprocess is enabled."
        ),
    )
    parser.add_argument(
        "--gpus",
        type=str,
        default=None,
        help=(
            "If subprocess is enabled, this is an optional comma separated "
            "list of GPU device IDs to assign to subprocesses via "
            "CUDA_VISIBLE_DEVICES. Each subprocess gets a single GPU from this"
            " pool; the pool also limits concurrency. You can oversubscribe "
            "GPUs by repeating device IDs, e.g. `0,0,1,1` to allow 2 "
            "subprocesses to share each GPU."
        ),
    )
    parser.add_argument(
        "--affinities",
        type=str,
        default=None,
        help=(
            "If subprocess is enabled, this is an optional comma separated "
            "list of affinities to use, one for each process. This ensures a "
            "single cpu core is used for each batch, regardless of other "
            "environment variables."
        ),
    )
    parser.add_argument(
        "--ray",
        action="store_true",
        help=(
            "Use a ray executor, either connecting to an existing cluster, "
            "or starting a new one with num_workers"
        ),
    )
    parser.add_argument(
        "--gpus-per-task",
        type=parse_int_or_float,
        default=None,
        help=(
            "The number of gpus to request per task, if using a ray executor. "
            "The overall GPUs available is set by CUDA_VISIBLE_DEVICES, which "
            "ray follows."
        ),
    )
    parser.add_argument(
        "--verbosity",
        type=int,
        default=1,
        help="The verbosity level.",
    )
    parser.add_argument(
        "--verbosity-grow",
        type=int,
        default=1,
        help="The verbosity level.",
    )
    args = parser.parse_args()

    parent_dir = Path(args.parent_dir).expanduser().resolve()

    # common thread control environment variables
    os.environ["OMP_NUM_THREADS"] = str(args.num_threads)
    os.environ["MKL_NUM_THREADS"] = str(args.num_threads)
    os.environ["OPENBLAS_NUM_THREADS"] = str(args.num_threads)
    os.environ["VECLIB_MAXIMUM_THREADS"] = str(args.num_threads)
    os.environ["NUMEXPR_NUM_THREADS"] = str(args.num_threads)
    os.environ["NUMBA_NUM_THREADS"] = str(args.num_threads)

    sys.path.append(str(parent_dir))
    import xyzpy

    grow_kwargs = {
        "num_workers": args.num_workers,
        "subprocess": args.subprocess,
        "raise_errors": args.raise_errors,
        "verbosity": args.verbosity,
        "verbosity_grow": args.verbosity_grow,
    }

    if args.subprocess:
        # this calls `xyzpy-grow` itself and we want to match env vars above
        grow_kwargs["num_threads"] = args.num_threads
        grow_kwargs["affinities"] = args.affinities
        grow_kwargs["gpus"] = args.gpus
        grow_kwargs["log"] = args.log

    if args.ray:
        if args.subprocess:
            raise xyzpy.utils.XYZError(
                "Cannot use subprocess mode with ray executor."
            )

        if args.gpus_per_task is None:
            grow_kwargs["executor"] = xyzpy.RayExecutor(
                num_cpus=args.num_workers,
            )
        else:
            grow_kwargs["executor"] = xyzpy.RayGPUExecutor(
                num_cpus=args.num_workers,
                gpus_per_task=args.gpus_per_task,
            )

    crop = xyzpy.Crop(name=args.crop_name, parent_dir=str(parent_dir))

    if not crop.is_prepared():
        raise xyzpy.utils.XYZError(f"The crop {crop} has not been sown yet.")

    if args.batch_ids == "missing":
        batch_ids = crop.missing_results()
    else:
        batch_ids = tuple(
            map(int, filter(None, args.batch_ids.replace(" ", "").split(",")))
        )

    print("Growing:")
    print(crop)
    crop.grow(batch_ids=batch_ids, **grow_kwargs)
    print("Done!")


if __name__ == "__main__":
    main()



import functools

import numpy as np
import pytest
from dask import delayed
from numpy.testing import assert_allclose

import xyzpy as xyz
from xyzpy.utils import _get_fn_name


class TestGetFnName:
    def test_normal(self):
        def foo(a, b):
            pass

        assert _get_fn_name(foo) == "foo"

    def test_partial(self):
        def foo(a, b):
            pass

        pfoo = functools.partial(foo, b=2)
        assert _get_fn_name(pfoo) == "foo"

    def test_delayed(self):
        @delayed
        def dfoo(a, b):
            pass

        assert _get_fn_name(dfoo) == "dfoo"


class TestProgbar:
    def test_normal(self):
        for i in xyz.progbar(range(10)):
            pass

    def test_overide_ascii(self):
        for i in xyz.progbar(range(10), ascii=False):
            pass


class TestTimer:
    def test_simple(self):
        import time

        with xyz.Timer() as timer:
            time.sleep(0.01)
        assert timer.t > 0.0


class TestBenchmark:
    def test_no_setup_no_size(self):
        t = xyz.benchmark(lambda: np.linalg.eig(np.random.randn(100, 100)))
        assert t > 0

    def test_no_setup_with_size(self):
        n = 10
        t = xyz.benchmark(lambda n: np.linalg.eig(np.random.randn(n, n)), n=n)
        assert t > 0

    def test_with_setup_no_size(self):
        t = xyz.benchmark(
            fn=lambda X: np.linalg.eig(X),
            setup=lambda: np.random.randn(10, 10),
        )
        assert t > 0

    def test_with_setup_with_size(self):
        t = xyz.benchmark(
            fn=lambda X: np.linalg.eig(X),
            setup=lambda n: np.random.randn(n, n),
            n=10,
        )
        assert t > 0


class TestBenchmarker:
    def test_basic(self):

        def add1(a, b):
            return a + b

        def add2(a, b):
            c = np.zeros_like(a)
            c += b
            return c

        kernels = [add1, add2]

        def setup(n):
            a = np.random.randn(n)
            b = np.random.randn(n)
            return a, b

        benchmark_opts = {"starmap": True, "min_t": 0.01, "repeats": 3}

        b = xyz.Benchmarker(kernels, setup, benchmark_opts=benchmark_opts)

        ns = [2**i for i in range(1, 6)]

        b.run(ns, verbosity=2)
        b.lineplot()
        b.ilineplot()


class TestRunningStatistics:
    def test_basic(self):
        rs = xyz.RunningStatistics()

        rs.update(42.0)
        rs.update(42.0)
        rs.update(42.0)
        rs.update(42.0)

        assert rs.mean == pytest.approx(42.0)
        assert rs.var == pytest.approx(0.0)
        assert rs.std == pytest.approx(0.0)
        assert rs.err == pytest.approx(0.0)

        rs.update_from_it([44.0, 44.0, 44.0, 44.0])

        assert rs.mean == pytest.approx(43.0)
        assert rs.std == pytest.approx(1.0)
        assert rs.rel_err == pytest.approx(1.0 / (43 * 8**0.5))


class TestRunningCovariance:
    def test_matches_numpy(self):
        xs = np.random.randn(100)
        ys = np.random.randn(100)
        rc = xyz.RunningCovariance()
        rc.update(xs[0], ys[0])
        rc.update(xs[1], ys[1])
        rc.update(xs[2], ys[2])
        assert rc.sample_covar == pytest.approx(np.cov(xs[:3], ys[:3])[0, 1])
        rc.update_from_it(xs[3:], ys[3:])
        assert rc.sample_covar == pytest.approx(np.cov(xs, ys)[0, 1])
        assert rc.covar == pytest.approx(np.cov(xs, ys, bias=True)[0, 1])

    def test_matrix_form(self):
        xs = np.random.rand(1000)
        ys = 0.2 * xs * np.random.rand(1000)
        zs = 0.4 * ys * np.random.rand(1000)
        rcm = xyz.RunningCovarianceMatrix(n=3)
        rcm.update_from_it(xs, ys, zs)
        assert_allclose(np.cov([xs, ys, zs]), rcm.sample_covar_matrix)
        assert_allclose(np.cov([xs, ys, zs], bias=True), rcm.covar_matrix)


class TestFromRepeats:
    def test_basic(self):

        def fn(n):
            return np.random.rand(n).sum()

        rs = xyz.estimate_from_repeats(fn, 10, verbosity=2)
        assert rs.mean == pytest.approx(5, rel=0.1)


class TestGetSizeOf:
    def test_nested_list(self):
        import sys

        obj1 = [[1]]
        obj2 = [[1 << 100 - 1]]
        assert sys.getsizeof(obj1) == sys.getsizeof(obj2)
        assert xyz.getsizeof(obj1) != xyz.getsizeof(obj2)



# TODO: function_str i.e. (a, b, c) -> (x[t], y[t,w], z)

import inspect

from ..utils import XYZError, isiterable


def _str_2_tuple(x):
    """Ensure `x` is at least a 1-tuple of str."""
    return (x,) if isinstance(x, str) else tuple(x)


def dictify(x):
    """Ensure `x` is a dict."""
    if isinstance(x, dict):
        return x
    elif x:
        return dict(x)
    else:
        return dict()


def parse_fn_args(fn, fn_args):
    """Parse function arguments into a tuple of argument names."""
    if fn_args is None:
        return tuple(inspect.signature(fn).parameters)

    return _str_2_tuple(fn_args)


# combo_runner -------------------------------------------------------------- #


def check_for_duplicates(arg, values):
    """Raise if duplicate values are provided for combo argument ``arg``."""
    if values is Ellipsis:
        return
    seen = set()
    for val in values:
        if val in seen:
            raise XYZError(f"Duplicate combo value for '{arg}': {val}")
        seen.add(val)


def parse_combos(combos) -> tuple[tuple[str, list], ...]:
    """Turn dicts and single tuples into proper form for combo runners."""
    if not combos:
        return ()
    if isinstance(combos, dict):
        combos = tuple(combos.items())
    elif isinstance(combos[0], str):
        combos = (combos,)

    combos = tuple(
        (arg, list(vals) if isiterable(vals) else vals) for arg, vals in combos
    )

    for arg, values in combos:
        check_for_duplicates(arg, values)

    return combos


def parse_combo_results(results, var_names):
    """Normalize combo results to a tuple when a single output is used."""
    if var_names is not None and (
        isinstance(var_names, str) or len(var_names) == 1
    ):
        results = (results,)
    return results


# case_runner --------------------------------------------------------------- #


def parse_cases(cases, fn_args=None):
    """Normalize cases into a tuple of dicts keyed by ``fn_args``."""
    if cases is None:
        return None

    # cases = {'a': 1, 'b': 2, 'c': 3} --> ({'a': 1, 'b': 2, 'c': 3},)
    if isinstance(cases, dict):
        return (cases,)

    cases = tuple(cases)
    if isinstance(cases[0], dict):
        # fn_args contained in cases already
        return cases

    if fn_args is None:
        raise TypeError(
            "`fn_args` must be provided when `cases` is not "
            "provided as a iterable of `dict`."
        )

    # e.g. if fn_args = ('a',) and cases = (1, 10, 100)
    #     we want cases --> ((1,), (10,), (100,))
    if isinstance(cases[0], str) or not isiterable(cases[0]):
        cases = tuple((c,) for c in cases)

    cases = tuple(dict(zip(fn_args, c)) for c in cases)
    return cases


def parse_case_results(results, var_names):
    """Normalize case results to tuples for single-output functions."""
    if isinstance(var_names, str) or len(var_names) == 1:
        results = tuple((r,) for r in results)
    return results


# common variable description ----------------------------------------------- #


def parse_var_names(var_names):
    """Normalize variable names to a tuple (or ``(None,)``)."""
    return (
        (None,)
        if var_names is None
        else (var_names,)
        if isinstance(var_names, str)
        else tuple(var_names)
    )


def parse_var_dims(var_dims, var_names):
    """Parse function mapping parameters into standard form.

    Parameters
    ----------
    var_dims : dict, tuple, or str
        * dict
            Mapping of each output to its dimensions, each either str
            or tuple of str. The keys themselves can be a tuple of several
            output names if they all have the same dimensions.
        * tuple
            List of output dimensions directly corresponding to list of
            var_names. Must be same length as `var_names`
        * str
            Only allowed for single output with single dimension.
    var_names : tuple of str, str, or None
        * tuple of str
            List of names of var_names.
        * str
            Single named output.
        * None
            Automatic result output using Dataset/DataArray, in this case
            check that var_dims is None as well.
    """
    if var_names is None:
        if var_dims is not None:
            raise ValueError(
                "Cannot specify variable dimensions if using"
                "automatic dataset output (var_names=None)."
            )
        return dict()

    new_var_dims = {k: () for k in var_names}  # default to empty tuple

    if not var_dims:
        return new_var_dims
    # check if single output with single dim
    elif isinstance(var_dims, str):
        if len(var_names) != 1:
            raise ValueError(
                "When `var_dims` is specified as a single "
                "string, there must be a single output, for "
                "which it is the single dimension."
            )
        var_dims = {var_names[0]: (var_dims,)}
    # check for direct correspondence to var_names
    elif isinstance(var_dims, (tuple, list)) and any(
        isinstance(x, str) or x[0] not in var_names for x in var_dims
    ):
        if len(var_dims) != len(var_names):
            raise ValueError(
                "`var_dims` cannot be interpreted as a "
                "mapping of `var_names` to their dimensions "
                "and is the wrong length to be in one to "
                "one correspondence."
            )
        var_dims = dict(zip(var_names, var_dims))
    # assume dict-like
    else:
        var_dims = dict(var_dims)

    # update new_var_dims, splitting var_names defined as having the same
    #   dims, and making sure all dims are tuple.
    try:
        for k, v in var_dims.items():
            v = (v,) if isinstance(v, str) else tuple(v)

            if isinstance(k, str):
                if k not in var_names:
                    raise KeyError

                new_var_dims[k] = v

            else:
                for sub_k in k:
                    if sub_k not in var_names:
                        raise KeyError

                    new_var_dims[sub_k] = v

    except KeyError:
        raise ValueError(
            "An unexpected output name was specified in the "
            "output dimensions mapping."
        )

    return new_var_dims


parse_var_coords = dictify
parse_constants = dictify
parse_resources = dictify
parse_attrs = dictify



""" """
# TODO: marker alpha

import functools
import itertools

import numpy as np

from ..manage import auto_xyz_ds
from .core import (
    PLOTTER_DEFAULTS,
    AbstractHeatMap,
    AbstractLinePlot,
    AbstractScatter,
    Plotter,
    calc_row_col_datasets,
    intercept_call_arg,
    prettify,
)


@functools.lru_cache(1)
def _init_bokeh_nb():
    """Cache this so it doesn't happen over and over again."""
    from bokeh.plotting import output_notebook
    from bokeh.resources import INLINE

    output_notebook(resources=INLINE)


def bshow(figs, nb=True, interactive=False, **kwargs):
    """ """
    from bokeh.plotting import show

    if nb:
        _init_bokeh_nb()
        show(figs, notebook_handle=interactive)
    else:
        show(figs)


# --------------------------------------------------------------------------- #
#                     Main lineplot interface for bokeh                       #
# --------------------------------------------------------------------------- #


class PlotterBokeh(Plotter):
    def __init__(self, ds, x, y, z=None, **kwargs):
        """ """
        # bokeh custom options / defaults
        kwargs["return_fig"] = kwargs.pop("return_fig", False)
        self._interactive = kwargs.pop("interactive", False)

        super().__init__(ds, x, y, z, **kwargs, backend="BOKEH")

    def prepare_axes(self):
        """Make the bokeh plot figure and set options."""
        from bokeh.plotting import figure

        if self.add_to_axes is not None:
            self._plot = self.add_to_axes

        else:
            # Currently axes scale type must be set at figure creation?
            self._plot = figure(
                # convert figsize to roughly matplotlib dimensions
                width=int(
                    self.figsize[0] * 80
                    + (100 if self._use_legend else 0)
                    + (20 if self._ytitle else 0)
                    + (20 if not self.yticklabels_hide else 0)
                ),
                height=int(
                    self.figsize[1] * 80
                    + (20 if self.title else 0)
                    + (20 if self._xtitle else 0)
                    + (20 if not self.xticklabels_hide else 0)
                ),
                x_axis_type=("log" if self.xlog else "linear"),
                y_axis_type=("log" if self.ylog else "linear"),
                y_axis_location=("right" if self.ytitle_right else "left"),
                title=self.title,
                toolbar_location="above",
                toolbar_sticky=False,
                active_scroll="wheel_zoom",
            )

    def set_axes_labels(self):
        """Set the labels on the axes."""
        if self._xtitle:
            self._plot.xaxis.axis_label = self._xtitle
        if self._ytitle:
            self._plot.yaxis.axis_label = self._ytitle

    def set_axes_range(self):
        """Set the plot ranges of the axes, and the panning limits."""
        from bokeh.models import DataRange1d

        self.calc_data_range()

        # plt_x_centre = (self._data_xmax + self._data_xmin) / 2
        # plt_x_range = self._data_xmax - self._data_xmin
        # xbounds = (plt_x_centre - plt_x_range, plt_x_centre + plt_x_range)
        xbounds = None
        self._plot.x_range = (
            DataRange1d(
                start=self._xlims[0], end=self._xlims[1], bounds=xbounds
            )
            if self._xlims
            else DataRange1d(bounds=xbounds)
        )

        # plt_y_centre = (self._data_ymax + self._data_ymin) / 2
        # plt_y_range = abs(self._data_ymax - self._data_ymin)
        # ybounds = (plt_y_centre - plt_y_range, plt_y_centre + plt_y_range)
        ybounds = None
        self._plot.y_range = (
            DataRange1d(
                start=self._ylims[0], end=self._ylims[1], bounds=ybounds
            )
            if self._ylims
            else DataRange1d(bounds=ybounds)
        )

    def set_spans(self):
        """Set custom horizontal and verical line spans."""
        from bokeh.models import Span

        span_opts = {
            "level": "glyph",
            "line_dash": "dashed",
            "line_color": (127, 127, 127),
            "line_width": self.span_width,
        }

        if self.hlines:
            for hl in self.hlines:
                self._plot.add_layout(
                    Span(location=hl, dimension="width", **span_opts)
                )
        if self.vlines:
            for vl in self.vlines:
                self._plot.add_layout(
                    Span(location=vl, dimension="height", **span_opts)
                )

    def set_gridlines(self):
        """Set whether to use gridlines or not."""
        if not self.gridlines:
            self._plot.xgrid.visible = False
            self._plot.ygrid.visible = False
        else:
            self._plot.xgrid.grid_line_dash = self.gridline_style
            self._plot.ygrid.grid_line_dash = self.gridline_style

    def set_tick_marks(self):
        """Set custom locations for the tick marks."""
        from bokeh.models import FixedTicker

        if self.xticks:
            self._plot.xaxis[0].ticker = FixedTicker(ticks=self.xticks)
        if self.yticks:
            self._plot.yaxis[0].ticker = FixedTicker(ticks=self.yticks)

        if self.xticklabels_hide:
            self._plot.xaxis.major_label_text_font_size = "0pt"
        if self.yticklabels_hide:
            self._plot.yaxis.major_label_text_font_size = "0pt"

    def set_sources_heatmap(self):
        from bokeh.plotting import ColumnDataSource

        # initialize empty source
        if not hasattr(self, "_source"):
            self._source = ColumnDataSource(data=dict())

        # remove mask from data -> not necessary soon? / convert to nan?
        var = np.ma.getdata(self._heatmap_var)

        self._source.add([var], "image")
        self._source.add([self._data_xmin], "x")
        self._source.add([self._data_ymin], "y")
        self._source.add([self._data_xmax - self._data_xmin], "dw")
        self._source.add([self._data_ymax - self._data_ymin], "dh")

    def set_sources(self):
        """Set the source dictionaries to be used by the plotter functions.
        This is seperate to allow interactive updates of the data only.
        """
        from bokeh.plotting import ColumnDataSource

        # check if heatmap
        if hasattr(self, "_heatmap_var"):
            return self.set_sources_heatmap()

        # 'copy' the zlabels iterator into src_zlbs
        self._zlbls, src_zlbs = itertools.tee(self._zlbls)

        # Initialise with empty dicts
        if not hasattr(self, "_sources"):
            self._sources = [
                ColumnDataSource(dict()) for _ in range(len(self._z_vals))
            ]

        # range through all data and update the sources
        for i, (zlabel, data) in enumerate(zip(src_zlbs, self._gen_xy())):
            self._sources[i].add(data["x"], "x")
            self._sources[i].add(data["y"], "y")
            self._sources[i].add([zlabel] * len(data["x"]), "z_coo")

            # check for color for scatter plot
            if "c" in data:
                self._sources[i].add(data["c"], "c")

            # check if should set y_err as well
            if "ye" in data:
                y_err_p = data["y"] + data["ye"]
                y_err_m = data["y"] - data["ye"]
                self._sources[i].add(
                    list(zip(data["x"], data["x"])), "y_err_xs"
                )
                self._sources[i].add(list(zip(y_err_p, y_err_m)), "y_err_ys")

            # check if should set x_err as well
            if "xe" in data:
                x_err_p = data["x"] + data["xe"]
                x_err_m = data["x"] - data["xe"]
                self._sources[i].add(
                    list(zip(data["y"], data["y"])), "x_err_ys"
                )
                self._sources[i].add(list(zip(x_err_p, x_err_m)), "x_err_xs")

    def plot_legend(self, legend_items=None):
        """Add a legend to the plot."""
        if self._use_legend:
            from bokeh.models import Legend

            loc = {"best": "top_left"}.get(self.legend_loc, self.legend_loc)
            where = {None: "right"}.get(self.legend_where, self.legend_where)

            # might be manually specified, e.g. from multiplot
            if legend_items is None:
                legend_items = self._lgnd_items

            lg = Legend(items=legend_items)
            lg.location = loc
            lg.click_policy = "hide"
            self._plot.add_layout(lg, where)

            # Don't repeatedly redraw legend
            self._use_legend = False

    def set_mappable(self):
        import matplotlib as plt
        from bokeh.models import LinearColorMapper, LogColorMapper

        mappr_fn = LogColorMapper if self.colormap_log else LinearColorMapper
        bokehpalette = [plt.colors.rgb2hex(m) for m in self.cmap(range(256))]

        self.mappable = mappr_fn(
            palette=bokehpalette, low=self._zmin, high=self._zmax
        )

    def plot_colorbar(self):
        if self._use_colorbar:
            where = {None: "right"}.get(self.legend_where, self.legend_where)

            from bokeh.models import BasicTicker, ColorBar, LogTicker

            ticker = LogTicker if self.colormap_log else BasicTicker
            color_bar = ColorBar(
                color_mapper=self.mappable,
                location=(0, 0),
                ticker=ticker(desired_num_ticks=6),
                title=self._ctitle,
            )
            self._plot.add_layout(color_bar, where)

    def set_tools(self):
        """Set which tools appear for the plot."""
        from bokeh.models import HoverTool

        tooltips = [
            (
                "({}, {})".format(
                    self.x_coo,
                    self.y_coo if isinstance(self.y_coo, str) else None,
                ),
                "(@x, @y)",
            ),
        ]
        if self.z_coo:
            tooltips.append((self.z_coo, "@z_coo"))

        self._plot.add_tools(HoverTool(tooltips=tooltips))

    def update(self):
        from bokeh.io import push_notebook

        self.set_sources()
        push_notebook()

    def show(self, **kwargs):
        """Show the produced figure."""
        if self.return_fig:
            return self._plot
        bshow(self._plot, **kwargs)
        return self

    def prepare_plot(self):
        self.prepare_axes()
        self.set_axes_labels()
        self.set_axes_range()
        self.set_spans()
        self.set_gridlines()
        self.set_tick_marks()
        self.set_sources()


def bokeh_multi_plot(fn):
    """Decorate a plotting function to plot a grid of values."""

    @functools.wraps(fn)
    def multi_plotter(ds, *args, row=None, col=None, link=False, **kwargs):

        if (row is None) and (col is None):
            return fn(ds, *args, **kwargs)

        # Set some global parameters
        p = fn(ds, *args, **kwargs, call=False)
        p.prepare_data_multi_grid()

        kwargs["xlims"] = kwargs.get("xlims", [p._data_xmin, p._data_xmax])
        kwargs["ylims"] = kwargs.get("ylims", [p._data_ymin, p._data_ymax])

        kwargs["vmin"] = kwargs.pop("vmin", p.vmin)
        kwargs["vmax"] = kwargs.pop("vmax", p.vmax)

        # split the dataset into its respective rows and columns
        ds_r_c, nrows, ncols = calc_row_col_datasets(ds, row=row, col=col)

        # intercept figsize as meaning *total* size for whole grid
        figsize = kwargs.pop("figsize", None)
        if figsize is None:
            av_n = (ncols + nrows) / 2
            figsize = (2 * (4 / av_n) ** 0.5, 2 * (4 / av_n) ** 0.5)
        else:
            figsize = (figsize[0] / ncols, figsize[1] / nrows)
        kwargs["figsize"] = figsize

        # intercept return_fig for the full grid and other options
        return_fig = kwargs.pop("return_fig", False)

        subplots = {}

        # range through rows and do subplots
        for i, ds_r in enumerate(ds_r_c):
            # range through columns
            for j, sub_ds in enumerate(ds_r):
                skws = {"legend": False, "colorbar": False}

                # if not last row
                if i != nrows - 1:
                    skws["xticklabels_hide"] = True
                    skws["xtitle"] = ""

                # if not first column
                if j != 0:
                    skws["yticklabels_hide"] = True
                    skws["ytitle"] = ""

                # label each column
                if (i == 0) and (col is not None):
                    col_val = prettify(ds[col].values[j])
                    skws["title"] = "{} = {}".format(col, col_val)
                    fx = "fontsize_xtitle"
                    skws["fontsize_title"] = kwargs.get(
                        fx, PLOTTER_DEFAULTS[fx]
                    )

                # label each row
                if (j == ncols - 1) and (row is not None):
                    skws["ytitle_right"] = True
                    row_val = prettify(ds[row].values[i])
                    skws["ytitle"] = "{} = {}".format(row, row_val)

                subplots[i, j] = fn(
                    sub_ds,
                    *args,
                    return_fig=True,
                    call=False,
                    **{**kwargs, **skws},
                )

        from bokeh.layouts import gridplot

        plts = [[subplots[i, j]() for j in range(ncols)] for i in range(nrows)]

        # link zooming and panning between all plots
        if link:
            x_range, y_range = plts[0][0].x_range, plts[0][0].y_range
            for i in range(nrows):
                for j in range(ncols):
                    plts[i][j].x_range = x_range
                    plts[i][j].y_range = y_range

        # the main grid
        p._plot = gridplot(plts)

        if p._use_legend or p._use_colorbar:
            from bokeh.layouts import row
            from bokeh.models import ColorBar, GlyphRenderer, Legend, Range1d

            # plot dummy using last sub_ds
            skws = {
                "title": "",
                "legend_loc": "center_left",
                "legend_where": "left",
            }
            lgren = fn(sub_ds, *args, return_fig=True, **{**kwargs, **skws})

            # remove all but legend, colorbar and glyph renderers
            lgren.renderers = [
                r
                for r in lgren.renderers
                if isinstance(r, (Legend, GlyphRenderer, ColorBar))
            ]
            lgren.toolbar_location = None
            lgren.outline_line_color = None

            # size it - this is pretty hacky at the moment
            lgren.width = 120
            lgren.height = int(80 * figsize[1] * nrows + 100)
            lgren.x_range = Range1d(0, 0)
            lgren.y_range = Range1d(0, 0)

            # append to the right of the gridplot
            p._plot = row([p._plot, lgren])

        if return_fig:
            return p._plot
        bshow(p._plot)

    return multi_plotter


class ILinePlot(PlotterBokeh, AbstractLinePlot):
    def __init__(self, ds, x, y, z=None, y_err=None, x_err=None, **kwargs):
        super().__init__(ds, x, y, z=z, y_err=y_err, x_err=x_err, **kwargs)

    def plot_lines(self):
        """Plot the data and a corresponding legend."""
        self._lgnd_items = []

        for src in self._sources:
            col = next(self._cols)
            zlabel = next(self._zlbls)
            legend_pics = []

            if self.lines:
                line = self._plot.line(
                    "x",
                    "y",
                    source=src,
                    color=col,
                    line_dash=next(self._lines),
                    line_width=next(self._lws) * 1.5,
                )
                legend_pics.append(line)

            if self.markers:
                marker = next(self._mrkrs)
                m = getattr(self._plot, marker)(
                    "x",
                    "y",
                    source=src,
                    name=zlabel,
                    color=col,
                    fill_alpha=0.5,
                    line_width=0.5,
                    size=self._marker_size,
                )
                legend_pics.append(m)

            # Check if errors specified as well
            if self.y_err:
                err = self._plot.multi_line(
                    xs="y_err_xs",
                    ys="y_err_ys",
                    source=src,
                    color=col,
                    line_width=self.errorbar_linewidth,
                )
                legend_pics.append(err)
            if self.x_err:
                err = self._plot.multi_line(
                    xs="x_err_xs",
                    ys="x_err_ys",
                    source=src,
                    color=col,
                    line_width=self.errorbar_linewidth,
                )
                legend_pics.append(err)

            # Add the names and styles of drawn lines for the legend
            self._lgnd_items.append((zlabel, legend_pics))

    def __call__(self):
        self.prepare_data_single()
        # Bokeh preparation
        self.prepare_plot()
        self.plot_lines()
        self.plot_legend()
        self.plot_colorbar()
        self.set_tools()
        return self.show(interactive=self._interactive)


@bokeh_multi_plot
@intercept_call_arg
def ilineplot(ds, x, y, z=None, y_err=None, x_err=None, **kwargs):
    """From  ``ds`` plot lines of ``y`` as a function of ``x``, optionally for
    varying ``z``. Interactive,

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Dimension to plot along the x-axis.
    y : str or tuple[str]
        Variable(s) to plot along the y-axis. If tuple, plot each of the
        variables - instead of ``z``.
    z : str, optional
        Dimension to plot into the page.
    y_err : str, optional
        Variable to plot as y-error.
    x_err : str, optional
        Variable to plot as x-error.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return ILinePlot(ds, x, y, z, y_err=y_err, x_err=x_err, **kwargs)


class AutoILinePlot(ILinePlot):
    """Interactive raw data multi-line plot."""

    def __init__(self, x, y_z, **lineplot_opts):
        ds = auto_xyz_ds(x, y_z)
        super().__init__(ds, "x", "y", z="z", **lineplot_opts)


def auto_ilineplot(x, y_z, **lineplot_opts):
    """Auto version of :func:`~xyzpy.ilineplot` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoILinePlot(x, y_z, **lineplot_opts)()


# --------------------------------------------------------------------------- #


class IScatter(PlotterBokeh, AbstractScatter):
    def __init__(self, ds, x, y, z=None, **kwargs):
        super().__init__(ds, x, y, z, **kwargs, markers=True)

    def plot_scatter(self):
        self._lgnd_items = []

        for src in self._sources:
            if "c" in src.column_names:
                col = {"field": "c", "transform": self.mappable}
            else:
                col = next(self._cols)
            marker = next(self._mrkrs)
            zlabel = next(self._zlbls)
            legend_pics = []

            m = getattr(self._plot, marker)(
                "x",
                "y",
                source=src,
                name=zlabel,
                color=col,
                fill_alpha=0.5,
                line_width=0.5,
                size=self._marker_size,
            )
            legend_pics.append(m)

            # Add the names and styles of drawn markers for the legend
            self._lgnd_items.append((zlabel, legend_pics))

    def __call__(self):
        self.prepare_data_single()
        # Bokeh preparation
        self.prepare_plot()
        self.plot_scatter()
        self.plot_legend()
        self.plot_colorbar()
        self.set_tools()
        return self.show(interactive=self._interactive)


@bokeh_multi_plot
@intercept_call_arg
def iscatter(ds, x, y, z=None, y_err=None, x_err=None, **kwargs):
    """From  ``ds`` plot a scatter of ``y`` against ``x``, optionally for
    varying ``z``. Interactive.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Quantity to plot along the x-axis.
    y : str or tuple[str]
        Quantity(s) to plot along the y-axis. If tuple, plot each of the
        variables - instead of ``z``.
    z : str, optional
        Dimension to plot into the page.
    y_err : str, optional
        Variable to plot as y-error.
    x_err : str, optional
        Variable to plot as x-error.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return IScatter(ds, x, y, z, y_err=y_err, x_err=x_err, **kwargs)


class AutoIScatter(IScatter):
    def __init__(self, x, y_z, **iscatter_opts):
        ds = auto_xyz_ds(x, y_z)
        super().__init__(ds, "x", "y", z="z", **iscatter_opts)


def auto_iscatter(x, y_z, **iscatter_opts):
    """Auto version of :func:`~xyzpy.iscatter` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoIScatter(x, y_z, **iscatter_opts)()


# --------------------------------------------------------------------------- #


_HEATMAP_ALT_DEFAULTS = (
    ("legend", False),
    ("colorbar", True),
    ("colormap", "inferno"),
    ("gridlines", False),
    ("padding", 0),
    ("figsize", (5, 5)),  # try to be square, maybe use aspect_ratio??
)


class IHeatMap(PlotterBokeh, AbstractHeatMap):
    def __init__(self, ds, x, y, z, **kwargs):
        # set some heatmap specific options
        for k, default in _HEATMAP_ALT_DEFAULTS:
            if k not in kwargs:
                kwargs[k] = default
        super().__init__(ds, x, y, z, **kwargs)

    def plot_heatmap(self):
        self.calc_color_norm()
        self._plot.image(
            image="image",
            x="x",
            y="y",
            dw="dw",
            dh="dh",
            source=self._source,
            color_mapper=self.mappable,
        )

    def __call__(self):
        # Core preparation
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_heatmap()
        self.plot_colorbar()
        self.set_tools()
        return self.show(interactive=self._interactive)


@bokeh_multi_plot
@intercept_call_arg
def iheatmap(ds, x, y, z, **kwargs):
    """From  ``ds`` plot variable ``z`` as a function of ``x`` and ``y`` using
    a 2D heatmap. Interactive,

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Dimension to plot along the x-axis.
    y : str
        Dimension to plot along the y-axis.
    z : str, optional
        Variable to plot as colormap.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return IHeatMap(ds, x, y, z, **kwargs)


class AutoIHeatMap(IHeatMap):
    def __init__(self, x, **iheatmap_opts):
        ds = auto_xyz_ds(x)
        super().__init__(ds, "y", "z", "x", **iheatmap_opts)


def auto_iheatmap(x, **iheatmap_opts):
    """Auto version of :func:`~xyzpy.iheatmap` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoIHeatMap(x, **iheatmap_opts)()



"""Objects for labelling and succesively running functions."""

import functools
import os
import shutil
import warnings
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from ..manage import load_df, load_ds, save_df, save_ds
from . import cropping
from .case_runner import case_runner_to_ds, parse_into_cases
from .combo_runner import combo_runner_to_ds
from .prepare import (
    XYZError,
    parse_attrs,
    parse_cases,
    parse_combos,
    parse_constants,
    parse_fn_args,
    parse_resources,
    parse_var_coords,
    parse_var_dims,
    parse_var_names,
)

# --------------------------------------------------------------------------- #
#                                   RUNNER                                    #
# --------------------------------------------------------------------------- #


class Runner(object):
    """Container class with all the information needed to systematically
    run a function over many parameters and capture the output in a dataset.

    Parameters
    ----------
    fn : callable
        Function that produces a single instance of a result.
    var_names : str, sequence of str, or None
        The ordered name(s) of the ouput variable(s) of `fn`. Set this
        explicitly to None if `fn` outputs already labelled data as a dict,
        :class:`~xarray.Dataset`, or :class:`~xarray.DataArray`.
    fn_args : str, or sequence of str, optional
        The ordered name(s) of the input arguments(s) of `fn`. This is only
        needed if the cases or combos supplied are not dict-like.
    var_dims : dict-like, optional
        Mapping of output variables to their named internal dimensions, can be
        the names of ``constants``.
    var_coords : dict-like, optional
        Mapping of output variables named internal dimensions to the actual
        values they take.
    constants : dict-like, optional
        Constants arguments to be supplied to `fn`. These can be used as
        'var_dims', and will be saved as coords if so, otherwise as attributes.
    resources : dict-like, optional
        Like `constants` but not saved to the the dataset, e.g. if very big.
    attrs : dict-like, optional
        Any other miscelleous information to be saved with the dataset.
    default_runner_settings
        These keyword arguments will be supplied as defaults to any runner.
    """

    def __init__(
        self,
        fn,
        var_names,
        fn_args=None,
        var_dims=None,
        var_coords=None,
        constants=None,
        resources=None,
        attrs=None,
        **default_runner_settings,
    ):
        self.fn = fn
        self._var_names = parse_var_names(var_names)
        self._fn_args = parse_fn_args(fn, fn_args)
        self._var_dims = parse_var_dims(var_dims, self._var_names)
        self._var_coords = parse_var_coords(var_coords)
        self._constants = parse_constants(constants)
        self._resources = parse_resources(resources)
        self._attrs = parse_attrs(attrs)
        self._last_ds = None
        self.default_runner_settings = default_runner_settings

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    # Attributes which should be parsed ------------------------------------- #

    def _get_fn_args(self):
        return self._fn_args

    def _set_fn_args(self, fn_args):
        self._fn_args = parse_fn_args(self.fn, fn_args)

    def _del_fn_args(self):
        self._fn_args = None

    fn_args = property(
        _get_fn_args,
        _set_fn_args,
        _del_fn_args,
        "List of the names of the arguments that the Runner's function takes.",
    )

    def _get_var_names(self):
        return self._var_names

    def _set_var_names(self, var_names):
        self._var_names = parse_var_names(var_names)

    def _del_var_names(self):
        self._var_names = None

    var_names = property(
        _get_var_names,
        _set_var_names,
        _del_var_names,
        "List of the names of the variables that the "
        "Runner's function produces.",
    )

    def _get_var_dims(self):
        return self._var_dims

    def _set_var_dims(self, var_dims, var_names=None):
        if var_names is None:
            var_names = self._var_names
        self._var_dims = parse_var_dims(var_dims, var_names)

    def _del_var_dims(self):
        self._var_dims = None

    var_dims = property(
        _get_var_dims,
        _set_var_dims,
        _del_var_dims,
        "Mapping of each output variable to its named dimensions",
    )

    def _get_var_coords(self):
        return self._var_coords

    def _set_var_coords(self, var_coords):
        self._var_coords = parse_var_coords(var_coords)

    def _del_var_coords(self):
        self._var_coords = None

    var_coords = property(
        _get_var_coords,
        _set_var_coords,
        _del_var_coords,
        "Mapping of each variable named dimension to its coordinate values.",
    )

    def _get_constants(self):
        return self._constants

    def _set_constants(self, constants):
        self._constants = parse_constants(constants)

    def _del_constants(self):
        self._constants = None

    constants = property(
        _get_constants,
        _set_constants,
        _del_constants,
        "Mapping of constant arguments supplied to the Runner's function.",
    )

    def _get_resources(self):
        return self._resources

    def _set_resources(self, resources):
        self._resources = parse_resources(resources)

    def _del_resources(self):
        self._resources = None

    resources = property(
        _get_resources,
        _set_resources,
        _del_resources,
        "Mapping of constant arguments supplied to the "
        "Runner's function that are *not* saved with the "
        "dataset.",
    )

    @property
    def last_ds(self):
        return self._last_ds

    # Running methods ------------------------------------------------------- #

    def run_combos(self, combos, constants=(), **runner_settings):
        """Run combos using the function map and save to dataset.

        Parameters
        ----------
        combos : dict_like[str, iterable]
            The values of each function argument with which to evaluate all
            combinations.
        constants : dict, optional
            Extra constant arguments for this run, repeated arguments will
            take precedence over stored constants but for this run only.
        runner_settings
            Keyword arguments supplied to :func:`~xyzpy.combo_runner`.
        """
        combos = parse_combos(combos)
        self._last_ds = combo_runner_to_ds(
            fn=self.fn,
            combos=combos,
            var_names=self._var_names,
            var_dims=self._var_dims,
            var_coords=self._var_coords,
            constants={**self._constants, **dict(constants)},
            resources=self._resources,
            attrs=self._attrs,
            parse=False,
            **{**self.default_runner_settings, **runner_settings},
        )
        return self._last_ds

    def run_cases(self, cases, constants=(), fn_args=None, **runner_settings):
        """Run cases using the function and save to dataset.

        Parameters
        ----------
        cases : sequence of mappings or tuples
            A sequence of cases.
        constants : dict (optional)
            Extra constant arguments for this run, repeated arguments will
            take precedence over stored constants but for this run only.
        runner_settings
            Supplied to :func:`~xyzpy.case_runner`.
        """
        if fn_args is None:
            fn_args = self._fn_args

        cases = parse_cases(cases, fn_args)

        self._last_ds = case_runner_to_ds(
            fn=self.fn,
            fn_args=fn_args,
            cases=cases,
            var_names=self._var_names,
            var_dims=self._var_dims,
            var_coords=self._var_coords,
            constants={**self._constants, **dict(constants)},
            resources=self._resources,
            attrs=self._attrs,
            parse=False,
            **{**self.default_runner_settings, **runner_settings},
        )
        return self._last_ds

    def Crop(
        self,
        name=None,
        parent_dir=None,
        save_fn=None,
        batchsize=None,
        num_batches=None,
    ):
        """Return a Crop instance with this runner, from which ``fn``
        will be set, and then combos can be sown, grown, and reaped into the
        ``Runner.last_ds``. See :class:`~xyzpy.Crop`.

        Returns
        -------
        Crop
        """
        return cropping.Crop(
            farmer=self,
            name=name,
            parent_dir=parent_dir,
            save_fn=save_fn,
            batchsize=batchsize,
            num_batches=num_batches,
        )

    def __repr__(self):
        string = "<xyzpy.Runner>\n"
        string += "    fn: {self.fn}\n"

        if self.fn_args is not None:
            string += "    fn_args: {self.fn_args}\n"

        string += "    var_names: {self.var_names}\n"

        if self.var_dims is not None:
            string += "    var_dims: {self.var_dims}\n"

        if self.constants:
            self._constants_list = list(self.constants)
            string += "    constants: {self._constants_list}\n"

        if self.resources:
            self._resources_list = list(self.resources)
            string += "    resources: {self._resources_list}\n"

        return string.format(self=self)


def label(
    var_names,
    fn_args=None,
    var_dims=None,
    var_coords=None,
    constants=None,
    resources=None,
    attrs=None,
    harvester=False,
    sampler=False,
    engine=None,
    **default_runner_settings,
):
    """Convenient decorator to automatically wrap a function as a
    :class:`~xyzpy.Runner` or :class:`~xyzpy.Harvester`.

    Parameters
    ----------
    var_names : str, sequence of str, or None
        The ordered name(s) of the ouput variable(s) of `fn`. Set this
        explicitly to None if `fn` outputs already labelled data as a `dict`,
        :class:`~xarray.Dataset`, or :class:`~xarray.DataArray`.
    fn_args : str, or sequence of str, optional
        The ordered name(s) of the input arguments(s) of `fn`. This is only
        needed if the cases or combos supplied are not dict-like.
    var_dims : dict-like, optional
        Mapping of output variables to their named internal dimensions, can be
        the names of ``constants``.
    var_coords : dict-like, optional
        Mapping of output variables named internal dimensions to the actual
        values they take.
    constants : dict-like, optional
        Constants arguments to be supplied to `fn`. These can be used as
        'var_dims', and will be saved as coords if so, otherwise as attributes.
    resources : dict-like, optional
        Like `constants` but not saved to the the dataset, e.g. if very big.
    attrs : dict-like, optional
        Any other miscelleous information to be saved with the dataset.
    harvester : bool or str, optional
        If ``True``, wrap the runner as a :class:`~xyzpy.Harvester`, if a
        string, create the harvester with that as the ``data_name``.
    default_runner_settings
        These keyword arguments will be supplied as defaults to any runner.

    Examples
    --------

    Declare a function as a runner directly::

        >>> import xyzpy as xyz

        >>> @xyz.label(var_names=['sum', 'diff'])
        ... def foo(x, y):
        ...     return x + y, x - y
        ...

        >>> foo
        <xyzpy.Runner>
            fn: <function foo at 0x7f1fd8e5b1e0>
            fn_args: ('x', 'y')
            var_names: ('sum', 'diff')
            var_dims: {'sum': (), 'diff': ()}

        >>> foo(1, 2)  # can still call it normally
        (3, -1)

    """
    if harvester and sampler:
        raise ValueError("Cannot be both a harvester and a sampler.")

    def wrapper(fn):

        r = Runner(
            fn,
            var_names,
            fn_args=fn_args,
            var_dims=var_dims,
            var_coords=var_coords,
            constants=constants,
            resources=resources,
            attrs=attrs,
            **default_runner_settings,
        )

        if harvester:
            if harvester is True:
                data_name = None
            else:
                data_name = harvester
            r = Harvester(r, data_name=data_name, engine=engine)

        if sampler:
            if sampler is True:
                data_name = None
            else:
                data_name = sampler
            r = Sampler(r, data_name=data_name, engine=engine)

        return functools.update_wrapper(r, fn)

    return wrapper


# --------------------------------------------------------------------------- #
#                                 HARVESTER                                   #
# --------------------------------------------------------------------------- #


class Harvester(object):
    """Container class for collecting and aggregating data to disk.

    Parameters
    ----------
    runner : Runner
        Performs the runs and describes the results.
    data_name : str, optional
        Base file path to save data to.
    chunks : int or dict, optional
        If not None, passed to xarray so that the full dataset is loaded and
        merged into with on-disk dask arrays.
    engine : str, optional
        Engine to use to save and load datasets.
    full_ds : xarray.Dataset, optional
        Initialize the Harvester with this dataset as the intitial full
        dataset.

    Members
    -------
    full_ds : xarray.Dataset
        Dataset containing all data harvested so far, by default synced to
        disk.
    last_ds : xarray.Dataset
        Dataset containing just the data from the last harvesting run.
    """

    def __init__(
        self,
        runner: Runner,
        data_name=None,
        chunks=None,
        engine="h5netcdf",
        full_ds=None,
    ):
        self.runner = runner
        self.data_name = data_name
        if engine is None:
            # allow None for default
            engine = "h5netcdf"
        self.engine = engine
        self.chunks = chunks
        self._full_ds = full_ds

    @property
    def fn(self):
        return self.runner.fn

    @fn.setter
    def fn(self, fn):
        self.runner.fn = fn

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    @property
    def last_ds(self):
        """Dataset containing the last runs' data."""
        return self.runner.last_ds

    def load_full_ds(self, chunks=None, engine=None):
        """Load the disk dataset into ``full_ds``.

        Parameters
        ----------
        chunks : int or dict, optional
            If not None, passed to xarray so that the full dataset is loaded
            and merged into with on-disk dask arrays.
        engine : str, optional
            Engine to use to save and load datasets.
        """
        if self.data_name is None:
            return

        if engine is None:
            engine = self.engine

        if chunks is None:
            chunks = self.chunks

        # Check file exists and can be written to
        if os.access(self.data_name, os.W_OK):
            self._full_ds = load_ds(
                self.data_name, engine=engine, chunks=chunks
            )

        # Do nothing if file does not exist at all
        elif not Path(self.data_name).is_file():  # pragma: no cover
            pass

        # Catch read-only errors etc.
        else:
            raise OSError(
                "The file '{}' exists but cannot be written to".format(
                    self.data_name
                )
            )

    @property
    def full_ds(self):
        """Dataset containing all saved runs."""
        if self._full_ds is None:
            self.load_full_ds()
        return self._full_ds

    def save_full_ds(self, new_full_ds=None, engine=None):
        """Save `full_ds` onto disk. The old file is moved and kept as a backup
        in case of errors when writing the new dataset to disk.

        Parameters
        ----------
        new_full_ds : xarray.Dataset, optional
            Save this dataset as the new full dataset, else use the current
            full datset.
        engine : str, optional
            Engine to use to save and load datasets.
        """
        from ..manage import auto_add_extension

        if self.data_name is None:
            raise XYZError(
                "You didn't set a ``data_name`` for this harvester "
                "to use for persistent storage."
            )

        if engine is None:
            engine = self.engine

        if new_full_ds is not None:
            if self._full_ds is not None:
                self._full_ds.close()

            self._full_ds = new_full_ds

        file_path = Path(auto_add_extension(self.data_name, engine))
        backup_path = Path(str(file_path) + ".bak")

        # first move existing file to backup
        if file_path.exists():
            if backup_path.exists():
                # delete any old backups
                if engine == "zarr":
                    shutil.rmtree(backup_path)
                else:
                    backup_path.unlink()
            file_path.rename(backup_path)

        try:
            save_ds(self._full_ds, self.data_name, engine=engine)
        except Exception:
            # restore backup on error
            if backup_path.exists():
                if file_path.exists():
                    if engine == "zarr":
                        shutil.rmtree(file_path)
                    else:
                        file_path.unlink()
                backup_path.rename(file_path)
            raise
        else:
            # successful save, delete backup
            if backup_path.exists():
                if engine == "zarr":
                    shutil.rmtree(backup_path)
                else:
                    backup_path.unlink()

    def delete_ds(self, backup=False):
        """Delete the on-disk dataset, optionally backing it up first."""
        from ..manage import auto_add_extension

        file_path = Path(auto_add_extension(self.data_name, self.engine))

        if backup:
            import datetime

            ts = "{:%Y%m%d-%H%M%S}".format(datetime.datetime.now())
            shutil.copy(file_path, str(file_path) + f".BAK-{ts}")

        if self._full_ds is not None:
            self._full_ds.close()

        if self.engine == "zarr":
            shutil.rmtree(file_path)
        else:
            file_path.unlink()

    def add_ds(
        self, new_ds, sync=True, overwrite=None, chunks=None, engine=None
    ):
        """Merge a new dataset into the in-memory full dataset.

        Parameters
        ----------
        new_ds : xr.Dataset or xr.DataArray
            Data to be merged into the full dataset.
        sync : bool, optional
            If True (default), load and save the disk dataset before
            and after merging in the new data.
        overwrite : {None, False, True}, optional
            How to combine data from the new run into the current full_ds:

            - ``None`` (default): attempt the merge and only raise if
              data conflicts.
            - ``True``: overwrite conflicting current data with
              that from the new dataset.
            - ``False``: drop any conflicting data from the new dataset.

        chunks : int or dict, optional
            If not None, passed to xarray so that the full dataset is loaded
            and merged into with on-disk dask arrays.
        engine : str, optional
            Engine to use to save and load datasets.
        """
        if isinstance(new_ds, xr.DataArray):
            new_ds = new_ds.to_dataset()

        # get default chunks if not overidden
        if chunks is None:
            chunks = self.chunks
        if chunks is not None:
            new_ds = new_ds.chunk(chunks)

        # only sync with disk if data name present
        sync_with_disk = sync and (self.data_name is not None)
        if sync_with_disk:
            self.load_full_ds(chunks=chunks, engine=engine)

        if self._full_ds is None:
            # No full ds yet, deep copy to maintain distinction between
            #   'full_ds' and 'last_ds'.
            new_full_ds = new_ds.copy(deep=True)

        else:
            # Overwrite with new data
            if overwrite is True:
                new_full_ds = new_ds.combine_first(self._full_ds)
            # Overwrite nothing
            elif overwrite is False:
                new_full_ds = self._full_ds.combine_first(new_ds)
            # Merge, raising error if the two datasets conflict
            else:
                new_full_ds = self._full_ds.merge(
                    new_ds,
                    join="outer",
                    compat="no_conflicts",
                )

        # we do this as if the string encoding of the old dataset is a smaller
        # type than the new one, strings will be silently truncated
        new_full_ds = new_full_ds.drop_encoding()

        if sync_with_disk:
            self.save_full_ds(new_full_ds, engine=engine)
        else:
            self._full_ds = new_full_ds

    def expand_dims(
        self,
        name,
        value,
        engine=None,
    ):
        """Add a new coordinate dimension with ``name`` and ``value``. The
        change is immediately synced with the on-disk dataset. Useful if you
        want to expand the parameter space along a previously constant
        argument.
        """
        new_ds = self.full_ds.expand_dims(name)
        new_ds.coords[name] = [value]
        if self.data_name is not None:
            self.save_full_ds(new_ds, engine=engine)
        else:
            self._full_ds = new_ds

    def drop_sel(
        self,
        labels=None,
        *,
        errors="raise",
        engine=None,
        **labels_kwargs,
    ):
        """Drop specific values of coordinates from this harvester and its
        dataset. See
        http://xarray.pydata.org/en/latest/generated/xarray.Dataset.drop_sel.html.
        The change is immediately synced with the on-disk dataset.
        Useful for tidying uneeded data points.
        """
        new_ds = self.full_ds.drop_sel(labels, errors=errors, **labels_kwargs)
        if self.data_name is not None:
            self.save_full_ds(new_ds, engine=engine)
        else:
            self._full_ds = new_ds

    def _maybe_expand_combos(self, combos):
        """Expand combos with ellipses into full coordinate values from the
        current full dataset."""
        return tuple(
            (key, self.full_ds.coords[key].values if values is ... else values)
            for key, values in parse_combos(combos)
        )

    def harvest_combos(
        self,
        combos,
        *,
        cases=None,
        missing_only=False,
        sync=True,
        overwrite=None,
        chunks=None,
        engine=None,
        **runner_settings,
    ):
        """Run combos, automatically merging into an on-disk dataset.

        Parameters
        ---------
        combos : dict_like[str, iterable]
            The combos to run. The only difference here is that you can supply
            an ellipse ``...``, meaning the all values for that coordinate will
            be loaded from the current full dataset.
        missing_only : bool, optional
            If True, only run combos that are not already present in the
            on-disk dataset.
        sync : bool, optional
            If True (default), load and save the disk dataset before
            and after merging in the new data.
        overwrite : {None, False, True}, optional

            - ``None`` (default): attempt the merge and only raise if
              data conflicts.
            - ``True``: overwrite any conflicting current data with that from
              the new dataset.
            - ``False``: drop any conflicting data from the new dataset.

        chunks : bool, optional
            If not None, passed passed to xarray so that the full dataset is
            loaded and merged into with on-disk dask arrays.
        engine : str, optional
            Engine to use to save and load datasets.
        runner_settings
            Supplied to :func:`~xyzpy.combo_runner`.
        """
        # possibly load full range of coordinate values from full dataset
        combos = self._maybe_expand_combos(combos)

        if missing_only:
            # convert into missing cases only
            cases = parse_into_cases(combos, cases, ds=self.full_ds)

            if len(cases) == 0:
                return

            # all information about combos is now in cases
            combos = {}

        # generate the new dataset!
        ds = self.runner.run_combos(combos, cases=cases, **runner_settings)

        # merge into full dataset
        self.add_ds(
            ds,
            sync=sync,
            overwrite=overwrite,
            chunks=chunks,
            engine=engine,
        )

    def harvest_cases(
        self,
        cases,
        *,
        sync=True,
        overwrite=None,
        chunks=None,
        engine=None,
        **runner_settings,
    ):
        """Run cases, automatically merging into an on-disk dataset.

        Parameters
        ---------
        cases : list of dict or tuple
            The cases to run.
        sync : bool, optional
            If True (default), load and save the disk dataset before
            and after merging in the new data.
        overwrite : {None, False, True}, optional
            What to do regarding clashes with old data:

            - ``None`` (default): attempt the merge and only raise if
              data conflicts.
            - ``True``: overwrite conflicting current data with
              that from the new dataset.
            - ``False``: drop any conflicting data from the new dataset.

        chunks : bool, optional
            If not None, passed passed to xarray so that the full dataset is
            loaded and merged into with on-disk dask arrays.
        engine : str, optional
            Engine to use to save and load datasets.
        runner_settings
            Supplied to :func:`~xyzpy.case_runner`.
        """
        ds = self.runner.run_cases(cases, **runner_settings)
        self.add_ds(
            ds, sync=sync, overwrite=overwrite, chunks=chunks, engine=engine
        )

    def Crop(
        self,
        name=None,
        parent_dir=None,
        save_fn=None,
        batchsize=None,
        num_batches=None,
    ):
        """Return a Crop instance with this Harvester, from which `fn`
        will be set, and then combos can be sown, grown, and reaped into the
        ``Harvester.full_ds``. See :class:`~xyzpy.Crop`.

        Returns
        -------
        Crop
        """
        return cropping.Crop(
            farmer=self,
            name=name,
            parent_dir=parent_dir,
            save_fn=save_fn,
            batchsize=batchsize,
            num_batches=num_batches,
        )

    def __repr__(self):
        string = (
            "<xyzpy.Harvester>\n"
            "Runner: {self.runner}"
            "Sync file -->\n"
            "    {self.data_name}    [{self.engine}]"
        )

        return string.format(self=self)

    def cultivate(
        self,
        combos=None,
        cases=None,
        constants=None,
        name=None,
        batchsize=None,
        num_batches=None,
        missing_only=True,
        shuffle=True,
        subprocess="auto",
        raise_errors=True,
        verbosity=1,
        on_existing="ask",
        on_error="ask",
        clean_up=None,
        **grow_kwargs,
    ):
        """Convenience method to run a full cycle of parsing combos into
        missing cases only, then persistently growing those cases, and finally
        merging the results into the full dataset.

        Parameters
        ----------
        combos : dict_like[str, iterable]
            The combos to run. The only difference here is that you can supply
            an ellipse ``...``, meaning the all values for that coordinate will
            be loaded from the current full dataset.
        cases : sequence of mappings or tuples, optional
            A sequence of (partial) individual settings to run. For each case,
            all settings given by combos will be generated.
        constants : dict, optional
            Extra constant arguments for this run.
        name : str, optional
            Name for the crop to be used for on-disk storage of batches,
            results and logs. You can use different names to grow results for
            the same dataset concurrently.
        batchsize : int, optional
            If given, the target number of cases to sow in each batch. This is
            computed from ``num_batches`` if not given and 1 if neither given.
        num_batches : int, optional
            If given, the target number of batches to sow. This is computed
            from ``batchsize`` if not given and 1 if neither given.
        missing_only : bool, optional
            If True (default), only run cases that are not already present in
            the on-disk dataset
        shuffle : bool, optional
            If True (default), shuffle the order of cases before sowing and
            growing. This can be a useful basic form of load balancing.
        subprocess : "auto" or bool, optional
            Whether to grow each batch in a fresh subprocess. This adds about
            1 second of overhead per batch, but allows the number of threads,
            cpu affinity and gpu assignment to be controlled. If "auto"
            (default) then subprocesses will be used if `num_threads`, `gpus`
            or `affinities` are specified. See `grow_subprocess` for details.
        raise_errors : bool, optional
            If True (default), raise any errors that occur during growing,
            otherwise just log them and continue with the next batch.
        verbosity : int, optional
            The level of logging to print during the sow/grow/reap process.
            0: no output, 1: progress bars, 2: progress bars with current
            setting postfixed.
        on_existing : {'ask', 'reap', 'delete', 'skip', 'raise'}, optional
            What to do if a crop with the same name already exists on
            disk. Default is ``'ask'`` (interactive prompt).
        on_error : {'ask', 'reap', 'delete', 'skip', 'raise'}, optional
            What to do if an error occurs during grow/reap. Default is
            ``'ask'`` (interactive prompt).
        clean_up : bool or None, optional
            Whether to delete the on-disk batch, result and log files after
            successfully reaping.
        grow_kwargs
            Keyword arguments to be supplied to the grow method of the Crop
            instances, e.g. for controlling subprocess behaviour. See
            :meth:`xyzpy.Crop.grow` for details.

        See Also
        --------
        Harvester.harvest_combos
        """
        crop = self.Crop(name)

        if crop.is_prepared():
            crop.handle_existing(action=on_existing)
            if crop.is_prepared():
                warnings.warn(
                    f"Crop {crop} is still prepared, skipping cultivate."
                )
                return self.full_ds

        # first map to cases, possibly parsing out existing data
        if missing_only:
            ds = self.full_ds
        else:
            ds = None
        combos = self._maybe_expand_combos(combos)
        cases = parse_into_cases(combos, ds=ds)

        if cases:
            # now write the cases to disk
            crop.sow_combos(
                combos=None,
                cases=cases,
                constants=constants,
                batchsize=batchsize,
                num_batches=num_batches,
                shuffle=shuffle,
                verbosity=verbosity,
            )
            try:
                if subprocess == "auto":
                    # XXX: detect num_threads, affinites, gpus
                    subprocess = False

                # start growing!
                crop.grow(
                    subprocess=subprocess,
                    raise_errors=raise_errors,
                    verbosity=verbosity,
                    **grow_kwargs,
                )

                # accumulate the grown results into the full dataset
                crop.reap(
                    overwrite=not missing_only,
                    clean_up=clean_up,
                    verbosity=verbosity,
                )

            except (Exception, KeyboardInterrupt) as e:
                msg = f"Grow/reap errored with {str(e)}"
                crop.handle_existing(
                    action=on_error,
                    msg=msg,
                    e=e,
                    overwrite=not missing_only,
                )

        return self.full_ds


def cultivate(
    fn,
    *,
    var_names=None,
    data_name=None,
    runner_opts=None,
    harvester_opts=None,
    combos=None,
    cases=None,
    constants=None,
    name=None,
    batchsize=None,
    num_batches=None,
    missing_only=True,
    shuffle=True,
    subprocess="auto",
    raise_errors=True,
    verbosity=1,
    on_existing="ask",
    on_error="ask",
    clean_up=None,
    **grow_kwargs,
):
    """Convenience function to run a full cycle of annotating a function,
    parsing combos into missing cases only, then persistently growing those
    cases, and finally merging the results into the full dataset.

    Parameters
    ----------
    fn : callable
        The function to run over combos and cases. This will be wrapped in
        a :class:`~xyzpy.Runner` and :class:`~xyzpy.Harvester` to perform the
        cultivation process. If `var_names` is None, it should return a `dict`,
        :class:`~xarray.Dataset` or :class:`~xarray.DataArray`.
    var_names : str, sequence of str, or None
        The ordered name(s) of the ouput variable(s) of `fn`. Set this
        explicitly to None if `fn` outputs already labelled data as a `dict`,
        :class:`~xarray.Dataset`, or :class:`~xarray.DataArray`.
    data_name : str, optional
        If given, the on-disk file to sync results with. If not set there will
        be no persistent results, since the harvester created in this
        functional interface is ephemeral.
    runner_opts : dict, optional
        Keyword arguments to be supplied to :class:`~xyzpy.Runner`.
    harvester_opts : dict, optional
        Keyword arguments to be supplied to :class:`~xyzpy.Harvester`.
    combos : dict_like[str, iterable]
        The combos to run. The only difference here is that you can supply
        an ellipse ``...``, meaning the all values for that coordinate will
        be loaded from the current full dataset.
    cases : sequence of mappings or tuples, optional
        A sequence of (partial) individual settings to run. For each case,
        all settings given by combos will be generated.
    constants : dict, optional
        Extra constant arguments for this run.
    name : str, optional
        Name for the crop to be used for on-disk storage of batches,
        results and logs. You can use different names to grow results for
        the same dataset concurrently.
    batchsize : int, optional
        If given, the target number of cases to sow in each batch. This is
        computed from ``num_batches`` if not given and 1 if neither given.
    num_batches : int, optional
        If given, the target number of batches to sow. This is computed
        from ``batchsize`` if not given and 1 if neither given.
    missing_only : bool, optional
        If True (default), only run cases that are not already present in
        the on-disk dataset
    shuffle : bool, optional
        If True (default), shuffle the order of cases before sowing and
        growing. This can be a useful basic form of load balancing.
    subprocess : "auto" or bool, optional
        Whether to grow each batch in a fresh subprocess. This adds about
        1 second of overhead per batch, but allows the number of threads,
        cpu affinity and gpu assignment to be controlled. If "auto"
        (default) then subprocesses will be used if `num_threads`, `gpus`
        or `affinities` are specified. See `grow_subprocess` for details.
    raise_errors : bool, optional
        If True (default), raise any errors that occur during growing,
        otherwise just log them and continue with the next batch.
    verbosity : int, optional
        The level of logging to print during the sow/grow/reap process.
        0: no output, 1: progress bars, 2: progress bars with current
        setting postfixed.
    on_existing : {'ask', 'reap', 'delete', 'skip', 'raise'}, optional
        What to do if a crop with the same name already exists on
        disk. Default is ``'ask'`` (interactive prompt).
    on_error : {'ask', 'reap', 'delete', 'skip', 'raise'}, optional
        What to do if an error occurs during grow/reap. Default is
        ``'ask'`` (interactive prompt).
    clean_up : bool or None, optional
        Whether to delete the on-disk batch, result and log files after
        successfully reaping.
    grow_kwargs
        Keyword arguments to be supplied to the grow method of the Crop
        instances, e.g. for controlling subprocess behaviour. See
        :meth:`xyzpy.Crop.grow` for details.

    See Also
    --------
    Harvester.cultivate
    """

    runner_opts = {} if runner_opts is None else dict(runner_opts)
    runner_opts.setdefault("var_names", var_names)

    harvester_opts = {} if harvester_opts is None else dict(harvester_opts)
    harvester_opts.setdefault("data_name", data_name)

    if harvester_opts.get("data_name") is None:
        warnings.warn(
            "Functional `cultivate` without `data_name` "
            "means no persistent storage of results. "
        )

    runner = Runner(fn, **runner_opts)
    harvester = Harvester(runner, **harvester_opts)

    return harvester.cultivate(
        combos=combos,
        cases=cases,
        constants=constants,
        name=name,
        batchsize=batchsize,
        num_batches=num_batches,
        missing_only=missing_only,
        shuffle=shuffle,
        subprocess=subprocess,
        raise_errors=raise_errors,
        verbosity=verbosity,
        on_existing=on_existing,
        on_error=on_error,
        clean_up=clean_up,
        **grow_kwargs,
    )


class Sampler:
    """Like a Harvester, but randomly samples combos and writes the table of
    results to a ``pandas.DataFrame``.

    Parameters
    ----------
    runner : xyzpy.Runner
        Runner describing a labelled function to run.
    data_name : str, optional
        If given, the on-disk file to sync results with.
    default_combos : dict_like[str, iterable], optional
        The default combos to sample from (which can be overridden).
    full_df : pandas.DataFrame, optional
        If given, use this dataframe as the initial 'full' data.
    engine : {'pickle', 'csv', 'json', 'hdf', ...}, optional
        How to save and load the on-disk dataframe. See
        :func:`~xyzpy.manage.load_df` and :func:`~xyzpy.manage.save_df`.

    Attributes
    ----------
    full_df : pandas.DataFrame
        Dataframe describing all data harvested so far.
    last_df : pandas.Dataframe
        Dataframe describing the data harvested on the previous run.
    """

    def __init__(
        self,
        runner,
        data_name=None,
        default_combos=None,
        full_df=None,
        engine="pickle",
    ):
        self.runner = runner
        self.data_name = data_name
        self.default_combos = (
            {} if default_combos is None else dict(default_combos)
        )
        self._full_df = full_df
        self._last_df = None
        if engine is None:
            # allow None for default
            engine = "pickle"
        self.engine = engine

    @property
    def fn(self):
        return self.runner.fn

    @fn.setter
    def fn(self, fn):
        self.runner.fn = fn

    def load_full_df(self, engine=None):
        """Load the on-disk full dataframe into memory."""
        if engine is None:
            engine = self.engine

        # Check file exists and can be written to
        if os.access(self.data_name, os.W_OK):
            self._full_df = load_df(self.data_name, engine=engine)

        # Do nothing if file does not exist at all
        elif not Path(self.data_name).is_file():  # pragma: no cover
            pass

        # Catch read-only errors etc.
        else:
            raise OSError(
                "The file '{}' exists but cannot be written to".format(
                    self.data_name
                )
            )

    @property
    def full_df(self):
        """The dataframe describing all data harvested so far."""
        if self._full_df is None:
            self.load_full_df()
        return self._full_df

    @property
    def last_df(self):
        """The dataframe describing the last set of data harvested."""
        return self._last_df

    def save_full_df(self, new_full_df=None, engine=None):
        """Save `full_df` onto disk.

        Parameters
        ----------
        new_full_df : pandas.DataFrame, optional
            Save this dataframe as the new full dataframe, else use the
            current ``full_df``.
        engine : str, optional
            Which engine to save the dataframe with, if None use the default.
        """
        if engine is None:
            engine = self.engine

        if new_full_df is not None:
            data_path = Path(self.data_name)
            if data_path.exists():
                data_path.unlink()
            self._full_df = new_full_df

        save_df(self._full_df, self.data_name, engine=engine)

    def delete_df(self, backup=False):
        """Delete the on-disk dataframe, optionally backing it up first."""
        if backup:
            import datetime

            ts = "{:%Y%m%d-%H%M%S}".format(datetime.datetime.now())
            shutil.copy(self.data_name, self.data_name + ".BAK-{}".format(ts))

        Path(self.data_name).unlink()

    def add_df(self, new_df, sync=True, engine=None):
        """Merge a new dataset into the in-memory full dataset.

        Parameters
        ----------
        new_df : pandas.DataFrame or dict
            Data to be appended to the full dataset.
        sync : bool, optional
            If True (default), load and save the disk dataframe before
            and after merging in the new data.
        engine : str, optional
            Which engine to save the dataframe with.
        """
        if isinstance(new_df, dict):
            new_df = pd.DataFrame(new_df)

        # only sync with disk if data name present
        sync_with_disk = sync and (self.data_name is not None)
        if sync_with_disk:
            self.load_full_df(engine=engine)

        if self._full_df is None:
            # No full df yet, deep copy to maintain distinction between
            #   'full_df' and 'last_df'.
            new_full_df = new_df.copy(deep=True)
        else:
            new_full_df = pd.concat(
                [self._full_df, new_df], ignore_index=True, sort=True
            )

        if sync_with_disk:
            self.save_full_df(new_full_df, engine=engine)
        else:
            self._full_df = new_full_df

    def gen_cases_fnargs(self, n, combos=None):
        """ """
        combos = {} if combos is None else dict(combos)
        combos = {**self.default_combos, **combos}
        cases = tuple(
            tuple(
                v() if callable(v) else np.random.choice(v)
                for v in combos.values()
            )
            for _ in range(n)
        )
        return tuple(combos.keys()), cases

    def sample_combos(
        self,
        n,
        combos=None,
        engine=None,
        **case_runner_settings,
    ):
        """Sample the target function many times, randomly choosing parameter
        combinations from ``combos`` (or ``SampleHarvester.default_combos``).

        Parameters
        ----------
        n : int
            How many samples to run.
        combos : dict_like[str, iterable], optional
            A mapping of function arguments to potential choices. Any keys in
            here will override ``default_combos``. You can also suppply a
            callable to manually return a random choice e.g. from a probability
            distribution.
        engine : str, optional
            Which method to use to sync with the on-disk dataframe.
        case_runner_settings
            Supplied to :func:`~xyzpy.case_runner` and so onto
            :func:`~xyzpy.combo_runner`. This includes ``parallel=True`` etc.
        """
        fn_args, cases = self.gen_cases_fnargs(n, combos)
        last_df = self.runner.run_cases(
            cases, fn_args=fn_args, to_df=True, **case_runner_settings
        )
        self._last_df = last_df
        self.add_df(last_df, engine=engine)
        return last_df

    def Crop(
        self,
        name=None,
        parent_dir=None,
        save_fn=None,
        batchsize=None,
        num_batches=None,
    ):
        """Return a Crop instance with this Sampler, from which `fn`
        will be set, and then samples can be sown, grown, and reaped into the
        ``Sampler.full_df``. See :class:`~xyzpy.Crop`.

        Returns
        -------
        Crop
        """
        return cropping.Crop(
            farmer=self,
            name=name,
            parent_dir=parent_dir,
            save_fn=save_fn,
            batchsize=batchsize,
            num_batches=num_batches,
        )

    def __repr__(self):
        string = (
            "<xyzpy.Sampler>\n"
            "Runner: {self.runner}"
            "Sync file -->\n"
            "    {self.data_name}    [{self.engine}]"
        )

        return string.format(self=self)



# FIXTURES ------------------------------------------------------------------ #


def foo3_scalar(a, b, c):
    assert abs(a) >= 0
    assert abs(a) < 10
    assert abs(b) >= 10
    assert abs(b) < 100
    assert abs(c) >= 100
    assert abs(c) < 1000
    return a + b + c


def foo3_float_bool(a, b, c):
    assert abs(a) >= 0
    assert abs(a) < 10
    assert abs(b) >= 10
    assert abs(b) < 100
    assert abs(c) >= 100
    assert abs(c) < 1000
    return a + b + c, a % 2 == 0


def foo2_scalar(a, b):
    assert abs(a) >= 0
    assert abs(a) < 10
    assert abs(b) >= 10
    assert abs(b) < 100
    return a + b


def foo2_array(a, b):
    assert abs(a) >= 0
    assert abs(a) < 10
    assert abs(b) >= 10
    assert abs(b) < 100
    return [b + a + 0.1 * i for i in range(10)]


def foo2_array_bool(a, b):
    assert abs(a) >= 0
    assert abs(a) < 10
    assert abs(b) >= 10
    assert abs(b) < 100
    return [b + a + 0.1 * i for i in range(10)], a % 2 == 0


def foo2_array_array(a, b):
    assert abs(a) >= 0
    assert abs(a) < 10
    assert abs(b) >= 10
    assert abs(b) < 100
    return ([b + i * a for i in range(5)], [b - i * a for i in range(5)])


def foo2_zarray1_zarray2(a, b):
    assert abs(a) >= 0
    assert abs(a) < 10
    assert abs(b) >= 10
    assert abs(b) < 100
    return (
        [b + a + 0.1j * i for i in range(5)],
        [b + a - 0.1j * i for i in range(5)],
    )


def foo_array_input(a, t):
    return tuple(a * x for x in t)


def foo2_dataset(a, b):
    import numpy as np
    import xarray as xr

    x = np.tile(a + b, (2, 3))
    return xr.Dataset({"x": (["t1", "t2"], x)})
